/************************************************************************
 * E.S.O. - VLT project_
 *
 * who       when      what
 * --------  --------  ----------------------------------------------
 * gduvert   2007-05-30 Updated the way diffVis Amp and Phase are computed.
 *                      This version should be closer to what was implemented
 *                      in yorick by F. Millour
 * gduvert   2007-03-20 Corrected the computation of the error on sigma2_V2
 *                      when errorType was THEORETICAL. We were using the
 *                      theoretical value of sigma2_P1P2 computed from the
 *                      retrieved instantaneous P1 and P2 photometries, but
 *                      not the corresponding averageP1P2, for which we used
 *                      the one stored in PiMultPj. But the latter is
 *                      P1*P2*sumvk, when the theoretical P1 and P2 are just
 *                      P1*sumv1k and P2*sumv2k.
 * gduvert   2007-03-11 Corrected the treatment of photometries at various
 *                      places: insured first that summations or means of
 *                      FluxSum- and FluxRat- and their associated Sigma2
 *                      values would be the FluxSum or FluxRat of (sum or
 *                      mean of per-beam photometries) and not the sum or mean
 *                      of (FluxSum or FluxRat). Second, introduced the 
 *                      tmpphot and s2phot temporary arrays to keep the 
 *                      per-beam fluxes and sigma2 values during all processes
 *                      (especially V2 debiassing) where these are needed.
 *                      Third, made apparent that sigma2FluxSumPiPj or 
 *                      sigma2FluxRatPiPj is not computed as should be
 *                      (using Papoulis development of a function of two
 *                      random variables) but are instead a coding of the
 *                      individual per-beam fluxes and errors. (We must
 *                      maintain these strange variables for backward
 *                      compatibility reasons, alas).
 * ealtariba 2006-12-04 fixed bug related to the use of p2vm's bad pixel and 
 *                      flat field maps.
 * gduvert   2006-11-23 - compute DiffVis wether frame selection is asked
 *                      or not
 * ealtariba 2006-09-24 - removed amdlibSelectFrames() static declaration
 *                      - fixed bug when doing threshold selection; tmpArray
 *                      was not initialized
 * gduvert   2006-07-04 - completed frame selection.
 *                      - added differential visibilities computation.
 * ealtariba 2006-05-24 added frame selection. Only frame selection on 
 *                      fringe constrast criteria at present time 
 *                      available.
 * ealtariba 2006-04-05 updated according to the new amdlibPISTON and
 *                      amdlibVIS_TABLE_ENTRY structures
 * altariba 2006-04-04 - differentiated piston and fringe constrast per band 
 *                       completed implementation related to station indexes and
 *                       projected baseline.
 * gzins      26/01/06 - forced theorical error estimation when binning = 1
 * gduvert    30/08/05 - sigma2Vis replaces visError.
 *                     - corrected sign bug (!) in amdlibMeanPiston.
 *                     - enforced use of amdlibBLANKING_VALUE.
 *                     - compute bispectrum errors.
 *                     - closure mean should avoid NaNs now.
 *                     - compilation of amdlibComputeCorrFluxDirect2T and
 *                       amdlibAverageSelectedVisibilities optional.
 * gduvert    28/08/05 - vis3Phi and vis3PhiErr in degrees, as requested
 *                       by the comment in amdlib.h (and OK with OI-FITS).
 * gduvert    22/08/05 - replaced DEBUG (used by ESO) by VERBOSE 
 *                       (private to amdlib)
 * gduvert    22/04/05 - corrected bug recently introduced in computation
 *                       of fringe criterion. Changed FC to a simpler
 *                       formula.
 * fmillour   21/04/05 - Cleaned the way the photometries are used :
 *                       now there is a new structure : amdlibPHOTOMETRY,
 *                       described by base and not by telescope.
 * gduvert    05/02/05 - removed all implicit allocations (non ISO C90).
 *                     - use of static variables in all programs defining the
 *                       amdlib.h structures (vis, vis2, etc) to avoid ugly
 *                       problems and complaints from vagrind/purify.
 *                     - use of amdlibBLANKING_VALUE;
 *                     - replace sigma2_ck by sigma2_CorrFlux;
 *                     - use exact error on photomery stored in Sigma2Phot1,etc
 *                       to add to sigma2_mk.
 *                     - added computation of mean piston bin by bin like for
 *                       the other measurements (implied by the OIFITS format)
 *                       ... And compute the corresponding "error" as the
 *                       jitter of those measurements.
 *                       ... Thus introduces a new "instantOpd" structure.
 *                     - insure that computePiston2T access the right data
 *                       when called from computePiston3T by passing nbBases 
 *                       and iBase.
 * gduvert    15/12/04 - cov_RI was not initialized in the THEORETICAL_ERROR
 *                       computation...
 * gduvert    10/12/04 - corrected computation of vis3 values (filled
 *                       the structure in fct of frame, closures and 
 *                       wavelengths)
 *                     - added amdlibAverageClosurePhases() to compute
 *                       1st order average of Closure Phases.
 *                     - added amdlibAverageSelectedVisibilities that
 *                       enable averaging of only the top N% of visibilities.
 *                     - corrected a buggy iFrame offset in 
 *                       amdlibCrossP2vmAndData
 *                     - retrieve and store all relevant configuration values
 *                       such as u,v 
 * gduvert    26/10/04 - delete estimate of RON of the detector from the
 *                       variance of what was left on the
 *                     - add estimation of a median sigma_det2 on the average 
 *                       of frames
 *                     - add computation of time series of sigma_det2
 *                     - add #if 0 and #if DEBUG to skip part of code
 * gduvert   03/03/03  derived from C++  routines written by Bram Acke
 *                     following yorick implementation by Eric Tatulli,
 *                     initially from documents DRSS and AMB-IGR-017 to
 *                     AMB-IGR-020 from various Authors, and the document
 *                     "Amber Data Reduction Software Specifications" by
 *                     G. Duvert. 
*/

/**
 * @file
 * Functions to compute the instantaneous correlated fluxes, pistons, 
 * photometries and fringe quality estimator, and to derive bin-averaged
 * values of normalised visibilities, debiassed squared visibilities, 
 * triple products. Additionally contains routines to estimate noise-averaged
 * values of these multidimensional values.
 */


static char *rcsId="@(#) $Id: amdlibVisibilities.c,v 1.5 2011/01/24 12:44:05 agabasch Exp $"; 
static void *use_rcsId = (0 ? &use_rcsId : (void*)&rcsId);

#define _POSIX_SOURCE 1

/*
 * System Headers
 */
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include "fitsio.h"

/*
 * Local Headers
 */
#include "amdlib.h"
#include "amdlibProtected.h"

/** the DATE part  of an ISO 8601 Date+Time format*/
#define amdlibDATE_LENGTH 10 

/** List of instrument configuration keywords which influence the P2VM */
static char *amdlibP2vmInsCfgKeywList[] = {
    "HIERARCH ESO INS OPTI2 NAME", 
    "HIERARCH ESO INS GRIS1 NAME",
    "HIERARCH ESO INS GRIS2 NAME", 
    "HIERARCH ESO INS GRAT1 NAME",
    "HIERARCH ESO INS GRAT1 ORDER", 
    "HIERARCH ESO INS GRAT1 WLEN",
    NULL};

/* Local functions */
double amdlibAbacusErrPhi(double x);

static amdlibCOMPL_STAT amdlibCrossP2vmAndData(/* Input */
                                               amdlibP2VM_MATRIX  *p2vm,
                                               amdlibSCIENCE_DATA *data,
                                               amdlibSCIENCE_DATA *sky,
                                               amdlibWAVEDATA     *waveData,
                                               int                firstFrame,
                                               int                nbFrames,
                                               int                *channelNo,
                                               int                nbChannels,
                                               /* Output */
                                               amdlibWAVELENGTH   *wave,
                                               amdlibPHOTOMETRY   *photometry,
                                               double             *p2vCkDk,
                                               unsigned char     *badPixelFlags,
                                               double             *mk,
                                               double             *sigma2Mk,
                                               amdlibERROR_MSG    errMsg);

static amdlibCOMPL_STAT amdlibInvertP2v(/* Input */
                                        double          *p2v,
                                        unsigned char   *badp2v,
                                        double          *sigma2_mk,
                                        int             binning,
                                        int             nbLVis,
                                        int             nbPix,
                                        int             nbBases,
                                        /* Output */
                                        double          *v2p,
                                        amdlibERROR_MSG errMsg);

static amdlibCOMPL_STAT amdlibComputeCorrFlux(/* Input */
                                              double                *v2p,
                                              double                *mk,
                                              double                *sigma2_mk,
                                              int                   nbFrames,
                                              int                   nbLVis,
                                              int                   nbPix,
                                              int                   nbBases,
                                              /* Output */
                                              amdlibVIS             *vis,
                                              amdlibERROR_MSG       errMsg);

static amdlibCOMPL_STAT amdlibMeanCpxVis(
                                         amdlibVIS         *instantCorrFlux,
                                         amdlibPHOTOMETRY  *photometry,
                                         int               iBin,
                                         amdlibVIS         *meanVis);

static amdlibCOMPL_STAT amdlibSumPhotometry(
                                            amdlibPHOTOMETRY *instantPhotometry,
                                            int               iBin,
                                            amdlibPHOTOMETRY  *sumPhotometry);

static void amdlibComputeFringeCriterion(/* Input */
                                         amdlibVIS       *instantCorrFlux,
                                         int             iBin,
                                         amdlibBAND      band,
                                         amdlibSELECTION *selection,
                                         /* Output */
                                         amdlibVIS       *vis);

static void amdlibAverageVis2(amdlibVIS2 *vis2);

static amdlibCOMPL_STAT amdlibComputeSqVis(/* Input */
                                           amdlibVIS          *instantCorrFlux,
                                           amdlibPHOTOMETRY   *photometry,
                                           int                iBin,
                                           amdlibBAND         band,
                                           amdlibSELECTION    *selection,
                                           amdlibERROR_TYPE   errorType,
                                           /* Output */
                                           amdlibVIS2         *vis2,
                                           amdlibERROR_MSG    errMsg);

static amdlibCOMPL_STAT amdlibComputeDiffVis(/* Input */
                                       amdlibVIS              *instantCorrFlux,
                                       amdlibWAVELENGTH       *wave,
                                       amdlibPISTON           *instantOpd,
                                       int                    iBin,
                                       amdlibBAND             band,
                                       amdlibSELECTION        *selection,
				       amdlibERROR_TYPE   errorType,
                                       /* Output */
                                       amdlibVIS              *vis,
                                       amdlibERROR_MSG        errMsg);

static amdlibCOMPL_STAT amdlibAverageVis(/* Input */
                                    amdlibVIS              *instantCorrFlux,
                                    int                    iBin,
                                    amdlibBAND             band,
                                    amdlibSELECTION        *selection,
                                    /* Output */
                                    amdlibVIS              *vis,
                                    amdlibERROR_MSG        errMsg);

static amdlibCOMPL_STAT amdlibGetThreshold(double          *array,
                                           int             arraySize,
                                           double          percentage,
                                           double          *threshold,
                                           amdlibERROR_MSG errMsg);

/*
 * Public functions
 */
/**
 * @warning obsolete method. Please use amdlibComputeVisibilities.
 * 
 * @param data structure containing the science observation (where the fringes 
 * are)
 * @param sky (optional) structure containing the SKY observation 
 * @param p2vm structure where P2VM is stored
 * @param waveData the wavelength definition structure
 * @param band the spectral band to select in the observations 
 * @param binning the number of consecutives frames in the science observation
 * that will be used to measure a single point of visibilties in output.
 * the total time span of the frames in a bin should be of the order of
 * the coherence time of the atmosphere, unless the observation has 
 * benefitted of a fringe tracker device. At Paranal coherence time are 
 * rarely more than a few Millisecond. 
 * @param errorType indicates wether the noise figures are estimated 
 * statistically from the sequence of photometries during the bin 
 * (amdlibSTATISTICAL_ERROR) or theoretically by the poisson error on N photons
 * (amdlibTHEORETICAL_ERROR). The latter is forced obviously when the binsize
 * is 1.
 * @param pistonType indicates wether the piston is to be measured by fitting
 * a slope in the phasors amdlibITERATIVE_PHASOR or in the differential phases
 * after dewrapping (amdlibUNWRAPPED_PHASE)
 * @param noCheckP2vmId forces amdlib to use without question the passed p2vm,
 * even if its magic number is not OK. Can happen if the P2VM has been 
 * observed AFTER the science observations. 
 * @param photometry output structure containing the (binned) photometry 
 * results.
 * @param vis structure containing the (binned) raw visibilities.
 * @param vis2 structure containing the (binned) squared visibilities.
 * @param vis3 structure containing the (binned) triple products.
 * @param wave structure containing the wavelengths corresponding to the other
 * output structures. 
 * @param opd structure containing the (binned) piston measurements.
 * @param errMsg error description message returned if function fails.
 *
 * @return 
 * the number of valid wavelengths for input band on successful completion.
 * Otherwise -1 is returned.
 */
int amdlibExtractVisibilities(/* Input */
                              amdlibSCIENCE_DATA     *data,
                              amdlibSCIENCE_DATA     *sky,
                              amdlibP2VM_MATRIX      *p2vm,
                              amdlibWAVEDATA         *waveData,
                              amdlibBAND             band,
                              int                    binning,
                              amdlibERROR_TYPE       errorType,
                              amdlibPISTON_ALGORITHM pistonType,
                              amdlibBOOLEAN         noCheckP2vmId,
                              /* Output */
                              amdlibPHOTOMETRY       *photometry,
                              amdlibVIS              *vis,
                              amdlibVIS2             *vis2,
                              amdlibVIS3             *vis3,
                              amdlibWAVELENGTH       *wave,
                              amdlibPISTON           *opd,
                              amdlibERROR_MSG        errMsg)
{
    amdlibCPT_VIS_OPTIONS visOpt = {binning, errorType, pistonType, 
        noCheckP2vmId, amdlibNO_FRAME_SEL, 1};

    amdlibLogTrace("amdlibExtractVisibilities()");

    amdlibLogWarning("amdlibExtractVisibilities is obsolete. Please use "
                     "amdlibComputeVisibilities");
    
    return (amdlibComputeVisibilities(data, sky, p2vm, waveData, band, &visOpt, 
                                      photometry, vis, vis2, vis3, wave,
                                      opd, errMsg));
}


/**
 * main entry point, gathers all strutures in input and output,
 * merely calls amdlibComputeVisibilitiesFor()
 *
 * @param data structure containing the science observation (where the fringes 
 * are)
 * @param sky (optional) structure containing the SKY observation 
 * @param p2vm structure where P2VM is stored
 * @param waveData the wavelength definition structure
 * @param band the spectral band to select in the observations 
 * @param visOptions information necessary for visibilities computation
 * (binning, noise and piston computation, requirement about p2vm use and frame
 * selection). 
 * @param photometry output structure containing the (binned) photometry 
 * results.
 * @param vis structure containing the (binned) raw visibilities.
 * @param vis2 structure containing the (binned) squared visibilities.
 * @param vis3 structure containing the (binned) triple products.
 * @param wave structure containing the wavelengths corresponding to the other
 * output structures. 
 * @param opd structure containing the (binned) piston measurements.
 * @param errMsg error description message returned if function fails.
 *
 * @return 
 * the number of valid wavelengths for input band on successful completion.
 * Otherwise -1 is returned.
 */
int amdlibComputeVisibilities(/* Input */
                              amdlibSCIENCE_DATA     *data,
                              amdlibSCIENCE_DATA     *sky,
                              amdlibP2VM_MATRIX      *p2vm,
                              amdlibWAVEDATA         *waveData,
                              amdlibBAND             band,
                              amdlibCPT_VIS_OPTIONS  *visOptions,
                              /* Output */
                              amdlibPHOTOMETRY       *photometry,
                              amdlibVIS              *vis,
                              amdlibVIS2             *vis2,
                              amdlibVIS3             *vis3,
                              amdlibWAVELENGTH       *wave,
                              amdlibPISTON           *opd,
                              amdlibERROR_MSG        errMsg)
{
    int nbChannelsInBand = 0;
    int channelNoForBand[amdlibNB_SPECTRAL_CHANNELS];
    amdlibSELECTION selectedFrames;
    int i, base;
    amdlibBOOLEAN noVis3 = amdlibFALSE;

    amdlibLogTrace("amdlibComputeVisibilities()");

    if ((visOptions->binSize != 1) && 
        (visOptions->frameSelectionType != amdlibNO_FRAME_SEL))
    {
        amdlibLogError("Binning AND frame selection are not compatible");
        return 0;
    }
    
    /* Search for the number of valid (considering p2vm flags) spectral channels
     * for specified band */
    if (amdlibGetChannelsInBand(p2vm, data, waveData, band, 
                                &nbChannelsInBand, 
                                channelNoForBand) != amdlibSUCCESS)
    {
        return -1;
    }

    if (nbChannelsInBand == 0)
    {
        return 0;
    }

    /* Compute visibilities for these valid spectral channels. */
    if (amdlibComputeVisibilitiesFor(data, sky, p2vm, waveData, 
                                     channelNoForBand, nbChannelsInBand, 
                                     visOptions->binSize, 
                                     visOptions->errorType, 
                                     visOptions->pistonType, 
                                     visOptions->noCheckP2vmId, 
                                     photometry, vis, vis2, vis3, 
                                     wave, opd, errMsg) != amdlibSUCCESS)
    {
        return -1;
    }

    /* If required, perform frame selection */
    if (visOptions->frameSelectionType != amdlibNO_FRAME_SEL)
    {
        if (amdlibAllocateSelection(&selectedFrames, vis->nbFrames,
                                    vis->nbBases, errMsg) != amdlibSUCCESS)
        {
            return -1;
        }

        /* Select 'good' frames depending on selection criterion */
        if (amdlibSelectFrames(vis, photometry, opd,
                               visOptions->frameSelectionType, 
                               visOptions->frameSelectionRatio, 
                               &selectedFrames, 
                               band, errMsg) != amdlibSUCCESS)
        {
            return -1;
        }

        /* Average visibilities, photometries and pistons on good frames */
        if (amdlibAverageVisibilities(photometry,
                                      vis, vis2, vis3, 
                                      opd, band, wave,
                                      &selectedFrames, errMsg) != amdlibSUCCESS)
        {
            return -1;
        }

        for (base = 0; base < vis->nbBases; base++)
        {
            if (selectedFrames.band[band].nbSelectedFrames[base] == 0)
            {
                /* No frames are selected for that base, so flag them */
                for (i=0; i < vis->nbFrames; i++)
                {
                    memset(vis->table[i*vis->nbBases+base].flag,
                           amdlibTRUE, 
                           vis->nbWlen * sizeof(amdlibBOOLEAN));
                    memset(vis2->table[i*vis2->nbBases+base].flag,
                           amdlibTRUE, 
                           vis2->nbWlen * sizeof(amdlibBOOLEAN));
                }
                noVis3 = amdlibTRUE;
            }
        }
        if ((vis3 != NULL) && (noVis3 == amdlibTRUE))
        {
            for (i=0; i < vis3->nbClosures * vis3->nbFrames; i++)
            {
                memset(vis3->table[i].flag, amdlibTRUE, 
                       vis3->nbWlen * sizeof(amdlibBOOLEAN));
            }
        }

        amdlibReleaseSelection(&selectedFrames);
    }

    return nbChannelsInBand;    
}

/**
 * Select frames according to frame selection criterion chosen.
 *
 * @param vis input visibility structure.
 * @param phot input photometry structure.
 * @param opd input piston structure.
 * @param frameSelectionType frame criterion to apply.
 * @param frameSelectionRatio percentage of good frames to keep or, directly,
 * threshold corresponding to the chosen criterion.
 * @param selectedFrames output structure containing a description of selected
 * frames.
 * @param band the current spectral band treated.
 * @param errMsg error description message returned if function fails.
 *
 * @return 
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.
 */
amdlibCOMPL_STAT amdlibSelectFrames(amdlibVIS             *vis,
                                    amdlibPHOTOMETRY      *phot,
                                    amdlibPISTON          *opd,
                                    amdlibFRAME_SELECTION frameSelectionType,
                                    double                frameSelectionRatio,
                                    amdlibSELECTION       *selectedFrames,
                                    amdlibBAND            band,
                                    amdlibERROR_MSG       errMsg)
{
    int iFrame, iBase, iWlen, index, index2;
    amdlibVIS_TABLE_ENTRY **visTablePtr = NULL;
    amdlibPHOTOMETRY_TABLE_ENTRY **photTablePtr = NULL;
    float                        **opdTablePtr = NULL;
    
    double *tmpArray[amdlibNBASELINE];
    double threshold[amdlibNBASELINE];
    amdlibBOOLEAN select[amdlibNBASELINE] = {amdlibTRUE, amdlibTRUE, amdlibTRUE};

    amdlibBOOLEAN lessThan = amdlibTRUE; 

    amdlibLogTrace("amdlibSelectFrames()");

    if ((frameSelectionType == amdlibFRG_CONTRAST_SEL_PCG) ||
        (frameSelectionType == amdlibFRG_CONTRAST_SEL_THR))
    {
        /* Wrap vis table */
        visTablePtr = 
            (amdlibVIS_TABLE_ENTRY **)amdlibWrap2DArray(vis->table,
                                            vis->nbBases, vis->nbFrames,
                                            sizeof(amdlibVIS_TABLE_ENTRY),
                                            errMsg);
        if (visTablePtr == NULL)
        {
            return amdlibFAILURE;
        }
        /* Allocate memory for temporary arrays */
        for (iBase = 0; iBase < vis->nbBases; iBase++)
        {
            tmpArray[iBase] = calloc(selectedFrames->band[band].nbSelectedFrames[iBase], 
                                    sizeof(double));
            if (tmpArray[iBase] == NULL)
            {
                amdlibSetErrMsg("Could not allocate memory for temporary buffer");
                return amdlibFAILURE;
            }
        }
        
        /* Loop on frames, copy fringe contrast snr in a temporary array for
         * each baseline ... */
        for (iBase = 0; iBase < vis->nbBases; iBase++)
        {
            index = 0;
            for (iFrame = 0; iFrame < vis->nbFrames; iFrame++)
            {
                if (selectedFrames->band[band].isSelectedPt[iBase][iFrame] == amdlibTRUE)
                {
                    tmpArray[iBase][index] = 
                        visTablePtr[iFrame][iBase].frgContrastSnrArray[band];
                    index++;
                }
            }
        }
        
	    /* Set threshold */
        if (frameSelectionType == amdlibFRG_CONTRAST_SEL_THR)
        {
            for (iBase = 0; iBase < vis->nbBases; iBase++)
            {
                threshold[iBase] = frameSelectionRatio;
            }
        }
        else
        {
            /* Get threshold associated to the frameSelectionRatio 
             * by sorting the temporary array */        
            for (iBase = 0; iBase < vis->nbBases; iBase++)
            {
                if (amdlibGetThreshold(tmpArray[iBase], 
                                       selectedFrames->band[band].nbSelectedFrames[iBase], 
                                       frameSelectionRatio, 
                                       &threshold[iBase], 
                                       errMsg) != amdlibSUCCESS)
                {
                    amdlibLogWarning("Could not get threshold associated to "
                                     "the frameSelectionRatio");
                    amdlibLogWarningDetail(errMsg);
                    amdlibLogWarning("Could not perform frame selection on "
                                     "baseline %d in band %c",
                                     iBase+1, amdlibBandNumToStr(band));
                    select[iBase] = amdlibFALSE; 
                }
            }
        }
        /* Loop on frames and stamp the good ones (criterion upper than 
         * threshold) */
        for (iBase = 0; iBase < vis->nbBases; iBase++)
        {
            if (select[iBase] == amdlibTRUE)
            {
                index = 0;
                for (iFrame = 0; iFrame < vis->nbFrames; iFrame++)
                {
                    if (selectedFrames->band[band].isSelectedPt[iBase][iFrame] == amdlibTRUE)
                    {
                        if (tmpArray[iBase][index] < threshold[iBase])
                        {
                            selectedFrames->band[band].isSelectedPt[iBase][iFrame] = amdlibFALSE;
                            selectedFrames->band[band].nbSelectedFrames[iBase]--;
                        }
                         
                        index++;
                    }
                }
            }
	    /* In the case no selection was possible, set all frames to false */
	    else
	    {
                for (iFrame = 0; iFrame < vis->nbFrames; iFrame++)
                {
		    selectedFrames->band[band].isSelectedPt[iBase][iFrame] = amdlibFALSE;
		}
		selectedFrames->band[band].nbSelectedFrames[iBase] = 0;
	    }
        }
        for (iBase = 0; iBase < vis->nbBases; iBase++)
        {
            free(tmpArray[iBase]);
        }
        amdlibFree2DArrayWrapping((void **)visTablePtr);
    }
    else if ((frameSelectionType == amdlibFLUX_SEL_PCG) ||
             (frameSelectionType == amdlibFLUX_SEL_THR))
    {
        /* Wrap phot table */
        photTablePtr = 
            (amdlibPHOTOMETRY_TABLE_ENTRY **)amdlibWrap2DArray(phot->table,
                                        phot->nbBases, phot->nbFrames,
                                        sizeof(amdlibPHOTOMETRY_TABLE_ENTRY),
                                        errMsg);
        if (photTablePtr == NULL)
        {
            return amdlibFAILURE;
        }
        /* Allocate memory for temporary arrays */
        for (iBase = 0; iBase < phot->nbBases; iBase++)
        {
            tmpArray[iBase] = calloc(selectedFrames->band[band].nbSelectedFrames[iBase], 
                                    sizeof(double));
            if (tmpArray[iBase] == NULL)
            {
                amdlibSetErrMsg("Could not allocate memory for temporary buffer");
                return amdlibFAILURE;
            }
        }

        /* Loop on frames, insert flux criterion in the temporary array 
         * for each baseline. Flux criterion is here 
         * sqrt(Pi*Pj)/signedsqrt(fluxPi+fluxPj+npix*RON^2)
         * (summed over spectrum = all lambdas) 
         */
        for (iBase = 0; iBase < phot->nbBases; iBase++)
        {
            index = 0;
            for (iFrame = 0; iFrame < phot->nbFrames; iFrame++)
            {
                if (selectedFrames->band[band].isSelectedPt[iBase][iFrame] == amdlibTRUE)
                {
		    /* FIXME: handle of not a number:
		       they should not appear here ! */
		    index2=0;
                    for (iWlen = 0; iWlen < phot->nbWlen; iWlen++)
                    {
			if(!isnan(photTablePtr[iFrame][iBase].sigma2FluxSumPiPj[iWlen]))
			{
			    tmpArray[iBase][index] += 
				amdlibSignedSqrt(photTablePtr[iFrame][iBase].PiMultPj[iWlen])
				/ amdlibSignedSqrt(photTablePtr[iFrame][iBase].sigma2FluxSumPiPj[iWlen]);
			    index2++;
			}
		    }
		    /* bugfix: sqrt(nbWlen) instead of nbWlen (changed by
		       index for nans handling) */
                    tmpArray[iBase][index] /= sqrt(index2);
                    index++;
                }
            }
        }

        
        /* If frameSelectionRatio is > 1, interpret as an absolute 
         * threshold ...*/
        if (frameSelectionType == amdlibFLUX_SEL_THR)
        {
            for (iBase = 0; iBase < phot->nbBases; iBase++)
            {
                threshold[iBase] = frameSelectionRatio;
            }
        }
        else
        {
            /* Sort these temporary arrays and get associated thresholds */
            for (iBase = 0; iBase < phot->nbBases; iBase++)
            {
                if (amdlibGetThreshold(tmpArray[iBase], 
                                       selectedFrames->band[band].nbSelectedFrames[iBase], 
                                       frameSelectionRatio, 
                                       &threshold[iBase], 
                                       errMsg) != amdlibSUCCESS)
                {
                    amdlibLogWarning("Could not get threshold associated to "
                                     "the frameSelectionRatio");
                    amdlibLogWarningDetail(errMsg);
                    amdlibLogWarning("Could not perform frame selection on "
                                     "baseline %d in band %c",
                                     iBase+1, amdlibBandNumToStr(band));
                    select[iBase] = amdlibFALSE; 
                }
            }
        }
        /* Loop on frames and stamp the good ones ( criterion upper than 
         * threshold) */
        for (iBase = 0; iBase < phot->nbBases; iBase++)
        {
            if (select[iBase] == amdlibTRUE)
            {
                index = 0;
                for (iFrame = 0; iFrame < phot->nbFrames; iFrame++)
                {
                    if (selectedFrames->band[band].isSelectedPt[iBase][iFrame] == amdlibTRUE)
                    {
                        if (tmpArray[iBase][index] < threshold[iBase])
                        {
                            selectedFrames->band[band].isSelectedPt[iBase][iFrame] = amdlibFALSE;
                            selectedFrames->band[band].nbSelectedFrames[iBase]--;
                        }
                        index++;
                    }
                }
            }
	    /* In the case no selection was possible, set all frames to false */
	    else
	    {
                for (iFrame = 0; iFrame < phot->nbFrames; iFrame++)
                {
		    selectedFrames->band[band].isSelectedPt[iBase][iFrame] = amdlibFALSE;
		}
		selectedFrames->band[band].nbSelectedFrames[iBase] = 0;
	    }
        }
        for (iBase = 0; iBase < phot->nbBases; iBase++)
        {
            free(tmpArray[iBase]);
        }
        amdlibFree2DArrayWrapping((void **)photTablePtr);
    }
    else if ((frameSelectionType == amdlibOPD_SEL_PCG) ||
             (frameSelectionType == amdlibOPD_SEL_THR))
    {
        lessThan = amdlibFALSE;

        /* Wrap opd table */
        opdTablePtr = amdlibWrap2DArrayFloat(opd->pistonOPDArray[band],
                                             opd->nbBases,
                                             opd->nbFrames, errMsg);
        if (opdTablePtr == NULL)
        {
            return amdlibFAILURE;
        }

        /* Allocate memory for temporary arrays */
        for (iBase = 0; iBase < opd->nbBases; iBase++)
        {
            tmpArray[iBase] = 
                calloc(selectedFrames->band[band].nbSelectedFrames[iBase], 
                       sizeof(double));
            if (tmpArray[iBase] == NULL)
            {
                amdlibSetErrMsg("Could not allocate memory for temporary buffer");
                return amdlibFAILURE;
            }
        }

        /* Loop on frames, copy piston in a temporary array for each baseline
         * ... */
        for (iBase = 0; iBase < opd->nbBases; iBase++)
        {
            index = 0;
            for (iFrame = 0; iFrame < opd->nbFrames; iFrame++)
            {
                if (selectedFrames->band[band].isSelectedPt[iBase][iFrame] == amdlibTRUE)
                {
                    tmpArray[iBase][index] = 1.0/
			(abs(opdTablePtr[iFrame][iBase])+
			 (abs(opdTablePtr[iFrame][iBase])==0));
                    index++;
                }
            }
        }

        /* Threshold case */
        if (frameSelectionType == amdlibOPD_SEL_THR)
        {
            for (iBase = 0; iBase < opd->nbBases; iBase++)
            {
                threshold[iBase] = 1.0/(frameSelectionRatio+
					(frameSelectionRatio==0));
            }
        }
        else
        {
            /* Sort these temporary arrays and get associated thresholds */
            for (iBase = 0; iBase < opd->nbBases; iBase++)
            {
                if (amdlibGetThreshold(tmpArray[iBase], 
                                       selectedFrames->band[band].nbSelectedFrames[iBase], 
                                       frameSelectionRatio, 
                                       &threshold[iBase], 
                                       errMsg) != amdlibSUCCESS)
                {
                    amdlibLogWarning("Could not get threshold associated to "
                                     "the frameSelectionRatio");
                    amdlibLogWarningDetail(errMsg);
                    amdlibLogWarning("Could not perform frame selection "
                                     "on baseline %d in band %c",
                                     iBase+1, amdlibBandNumToStr(band));
                    select[iBase] = amdlibFALSE; 
                }
            }
        }
        /* Loop on frames and stamp the good ones (criterion upper than 
         * threshold) */
        for (iBase = 0; iBase < opd->nbBases; iBase++)
        {
            if (select[iBase] == amdlibTRUE)
            {
                index = 0;
                for (iFrame = 0; iFrame < opd->nbFrames; iFrame++)
                {
                    if (selectedFrames->band[band].isSelectedPt[iBase][iFrame] == amdlibTRUE)
                    {
                        if (tmpArray[iBase][index] < threshold[iBase])
                        {
                            selectedFrames->band[band].isSelectedPt[iBase][iFrame] = amdlibFALSE;
                            selectedFrames->band[band].nbSelectedFrames[iBase]--;
                        }
                        index++;
                    }
                }
            }
	    /* In the case no selection was possible, set all frames to false */
	    else
	    {
                for (iFrame = 0; iFrame < opd->nbFrames; iFrame++)
                {
		    selectedFrames->band[band].isSelectedPt[iBase][iFrame] = amdlibFALSE;
		}
		selectedFrames->band[band].nbSelectedFrames[iBase] = 0;
	    }
        }
        for (iBase = 0; iBase < opd->nbBases; iBase++)
        {
            free(tmpArray[iBase]);
        }
        amdlibFree2DArrayWrapping((void **)photTablePtr);
    }
    else if (frameSelectionType == amdlibNO_FRAME_SEL)
    {
        return amdlibSUCCESS;
    }
    else 
    {
        amdlibSetErrMsg("Invalid frame selection type '%d'",
                        frameSelectionType);
        return amdlibFAILURE;
    }

    /* Update frame selection */
    amdlibUpdateSelection(selectedFrames);

    return amdlibSUCCESS;
}

/**
 * Computes the selection threshold, ie the minimal possible value for frame 
 * selection criterion so as to a frame should be selected. 
 *
 * @param array input data 
 * @param arraySize size of array.
 * @param frameSelectionRatio percentage of good frames to keep.
 * @param threshold output threshold.
 * @param errMsg error description message returned if function fails.
 *
 * @return
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.
 */
amdlibCOMPL_STAT amdlibGetThreshold(double          *array,
                                    int             arraySize,
                                    double          frameSelectionRatio,
                                    double          *threshold,
                                    amdlibERROR_MSG errMsg)
{
    int    i, j;
    double val;
    int    index;
    amdlibBOOLEAN isSorted;
    double * phantomArray;

    phantomArray = calloc(arraySize, sizeof(double));
    memcpy(phantomArray, array, arraySize*sizeof(double));
    /* Protected Bubble sort, leaves array invariant! */
    isSorted = amdlibFALSE;
    i = 0; 
    while ((i < arraySize) && (isSorted == amdlibFALSE))
    {
        /* Assume array is already sorted */
        isSorted = amdlibTRUE;
        for (j = 1; j < arraySize - i; j++)
        {
            if (phantomArray[j] < phantomArray[j-1])
            {
                /* Invert both values */
                val = phantomArray[j-1];
                phantomArray[j-1] = phantomArray[j];
                phantomArray[j] = val;

                /* phantomArray is not sorted yet */
                isSorted = amdlibFALSE;
            }
        }
        i++;
    }

    /* Find threshold */
    index = floor((1.0-frameSelectionRatio)*arraySize);
    *threshold = phantomArray[index];
    free(phantomArray);

    if (isnan(*threshold))
    {
        amdlibSetErrMsg("Impossible to determine threshold; probably "
                        "due to poor data quality");
        return amdlibFAILURE;
    }

    return amdlibSUCCESS;
}

/** Useful macro to free amdlibAverageVisibilities structures */ 
#define amdlibAverageVisibilities_FREEALL()                   \
    amdlibFree2DArrayWrapping((void **)photTabPtr);   \
    amdlibFree2DArrayWrapping((void **)visTabPtr);    \
    amdlibFree2DArrayWrapping((void **)vis2TabPtr);   \
    amdlibFree2DArrayWrapping((void **)vis3TabPtr); \
    amdlibFree3DArrayDouble(tmpphot);                 \
    amdlibFree3DArrayDouble(s2phot);                  \
/**
 * Average data for selected frames in each output data structure for
 * visibilities computation. 
 *
 * @param phot input/output photometry structure. 
 * @param vis input/output OI_VIS structure.
 * @param vis2 input/output OI_VIS2 structure.
 * @param vis3 input/output OI_VIS3 structure.
 * @param opd input/output piston structure.
 * @param band the current spectral band treated.
 * @param wave structure containing the wavelengths. 
 * @param selectedFrames structure containing information about selected frames.
 * @param errMsg error description message returned if function fails.
 *
 * @return
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.
 */
amdlibCOMPL_STAT amdlibAverageVisibilities(amdlibPHOTOMETRY *phot,
                                           amdlibVIS        *vis, 
                                           amdlibVIS2       *vis2, 
                                           amdlibVIS3       *vis3,
                                           amdlibPISTON     *opd,
                                           amdlibBAND       band,
                                           amdlibWAVELENGTH *wave,
                                           amdlibSELECTION  *selectedFrames,
                                           amdlibERROR_MSG  errMsg)
{
    amdlibPHOTOMETRY_TABLE_ENTRY **photTabPtr = NULL;
    amdlibVIS_TABLE_ENTRY **visTabPtr = NULL;
    amdlibVIS2_TABLE_ENTRY **vis2TabPtr = NULL;
    amdlibVIS3_TABLE_ENTRY **vis3TabPtr = NULL;
    int nbFrames = vis->nbFrames;
    int nbBases = vis->nbBases;
    int nbWlen = vis->nbWlen;
    int i, j, k;
    int totalNbSel;

    /* Intermediate structures */
    amdlibPHOTOMETRY imdPhot = {NULL};
    amdlibVIS imdVis = {NULL};
    amdlibVIS2 imdVis2 = {NULL};
    amdlibVIS3 imdVis3 = {NULL};
    amdlibPISTON imdOpd = {NULL};

    double ***tmpphot = NULL, ***s2phot = NULL;
    double P1P2, P1MoreP2, P1DiviP2, P1, P2, sigma2P1MoreP2, sigma2P1DiviP2, 
           sigma2_P1, sigma2_P2;
    
    amdlibLogTrace("amdlibAverageVisibilities()");

    /* Allocate intermediate structures  which will contain mean values*/
    if (amdlibAllocatePhotometry(&imdPhot, 1, phot->nbBases, 
                                 phot->nbWlen) != amdlibSUCCESS)
    {
        amdlibAverageVisibilities_FREEALL();
        amdlibSetErrMsg("Could not allocate memory for intermediate structures");
        return amdlibFAILURE;
    }
    if (amdlibAllocateVis(&imdVis, 1, vis->nbBases, 
                          vis->nbWlen) != amdlibSUCCESS)
    {
        amdlibAverageVisibilities_FREEALL();
        amdlibReleasePhotometry(&imdPhot);
        amdlibSetErrMsg("Could not allocate memory for intermediate structures");
        return amdlibFAILURE;
    }
    if (amdlibAllocateVis2(&imdVis2, 1, vis2->nbBases, 
                           vis2->nbWlen) != amdlibSUCCESS)
    {
        amdlibAverageVisibilities_FREEALL();
        amdlibReleasePhotometry(&imdPhot);
        amdlibReleaseVis(&imdVis);
        amdlibSetErrMsg("Could not allocate memory for intermediate structures");
        return amdlibFAILURE;
    }

    if (nbBases > 1)
    {
        if (amdlibAllocateVis3(&imdVis3, 1, vis3->nbClosures, 
                               vis3->nbWlen) != amdlibSUCCESS)
        {
            amdlibAverageVisibilities_FREEALL();
            amdlibReleasePhotometry(&imdPhot);
            amdlibReleaseVis(&imdVis);
            amdlibReleaseVis2(&imdVis2);
            amdlibSetErrMsg("Could not allocate memory for intermediate structures");
            return amdlibFAILURE;
        }
    }
    if (amdlibAllocatePiston(&imdOpd, 1, opd->nbBases) != amdlibSUCCESS)
    {
        amdlibAverageVisibilities_FREEALL();
        amdlibReleasePhotometry(&imdPhot);
        amdlibReleaseVis(&imdVis);
        amdlibReleaseVis2(&imdVis2);
        amdlibReleaseVis3(&imdVis3);
        amdlibSetErrMsg("Could not allocate memory for intermediate structures");
        return amdlibFAILURE;
    }
    
    /* Allocate memory */
    tmpphot = amdlibAlloc3DArrayDouble(2, nbWlen, nbBases, errMsg);
    if (tmpphot == NULL)
    {
        amdlibAverageVisibilities_FREEALL();
        return amdlibFAILURE;
    }
    s2phot = amdlibAlloc3DArrayDouble(2, nbWlen, nbBases, errMsg);
    if (s2phot == NULL)
    {
        amdlibAverageVisibilities_FREEALL();
        return amdlibFAILURE;
    }

    /* Wrap on structures to average */
    photTabPtr = (amdlibPHOTOMETRY_TABLE_ENTRY **)
        amdlibWrap2DArray(phot->table,  nbBases, nbFrames, 
                          sizeof(*(phot->table)), errMsg);
    if (photTabPtr == NULL)
    {
        amdlibAverageVisibilities_FREEALL();
        return amdlibFAILURE;
    }
    visTabPtr = (amdlibVIS_TABLE_ENTRY **)
        amdlibWrap2DArray(vis->table, nbBases, nbFrames,
                          sizeof(*(vis->table)), errMsg);
    if (visTabPtr == NULL)
    {
        amdlibAverageVisibilities_FREEALL();
        return amdlibFAILURE;
    }
    vis2TabPtr = (amdlibVIS2_TABLE_ENTRY **)
        amdlibWrap2DArray(vis2->table, nbBases, nbFrames,
                          sizeof(*(vis2->table)), errMsg);
    if (vis2TabPtr == NULL)
    {
        amdlibAverageVisibilities_FREEALL();
        return amdlibFAILURE;
    }
    if (nbBases > 1)
    {
        vis3TabPtr = (amdlibVIS3_TABLE_ENTRY **)
            amdlibWrap2DArray(vis3->table, vis3->nbClosures, nbFrames,
                              sizeof(*(vis3->table)), errMsg);
        if (vis3TabPtr == NULL)
        {
            amdlibAverageVisibilities_FREEALL();
            return amdlibFAILURE;
        }
    }

    /* Check there is at least one frame selected */
    totalNbSel = 0;
    for (j=0; j < nbBases; j++)
    {
        totalNbSel += selectedFrames->band[band].nbSelectedFrames[j];
    }
    
    if (totalNbSel != 0)
    {
        /* Sum all values that have to be simply averaged */
        for (j = 0; j < nbBases; j++)
        {
            selectedFrames->band[band].firstSelectedFrame[j] = -1;
        }



        for (j = 0; j < nbBases; j++)
        {
            if (selectedFrames->band[band].nbSelectedFrames[j] == 0)
            {
                continue;
            }
            for (i = 0; i < nbFrames; i++)
            {
                if (selectedFrames->band[band].isSelectedPt[j][i] == amdlibTRUE)
                {
                    imdVis.table[j].expTime += visTabPtr[i][j].expTime;
                    imdVis2.table[j].expTime += vis2TabPtr[i][j].expTime;

                    for (k = 0; k < nbWlen; k++)
                    {
                        P1MoreP2 = photTabPtr[i][j].fluxSumPiPj[k];

                        sigma2P1MoreP2 = photTabPtr[i][j].sigma2FluxSumPiPj[k];

                        P1DiviP2 = photTabPtr[i][j].fluxRatPiPj[k];

                        sigma2P1DiviP2 = photTabPtr[i][j].sigma2FluxRatPiPj[k];

                        P2 = P1MoreP2 / (1.0 + P1DiviP2);

                        sigma2_P2 = sigma2P1MoreP2 / (1.0 + sigma2P1DiviP2);

                        P1 = P1MoreP2*P1DiviP2/(1.0+P1DiviP2);

                        sigma2_P1 = sigma2P1MoreP2 * sigma2P1DiviP2 / 
                            (1.0 + sigma2P1DiviP2);

                        tmpphot[j][k][0] += P1;
                        s2phot[j][k][0] += sigma2_P1;
                        tmpphot[j][k][1] += P2;
                        s2phot[j][k][1] += sigma2_P2;

                        imdPhot.table[j].PiMultPj[k] +=
                            photTabPtr[i][j].PiMultPj[k];
                    }

                    if (selectedFrames->band[band].firstSelectedFrame[j] == -1)
                    {
                        selectedFrames->band[band].firstSelectedFrame[j] = i;
                    }
                }
            }
        }

        for (j = 0; j < nbBases; j++)
        {
            if (selectedFrames->band[band].nbSelectedFrames[j] == 0)
            {
                continue;
            }
            for (k = 0; k < nbWlen; k++)
            {
                imdPhot.table[j].fluxSumPiPj[k] = tmpphot[j][k][0] + 
                    tmpphot[j][k][1];

                imdPhot.table[j].sigma2FluxSumPiPj[k] = s2phot[j][k][0] + 
                    s2phot[j][k][1];

                imdPhot.table[j].fluxRatPiPj[k] = tmpphot[j][k][0] / 
                    tmpphot[j][k][1];

                imdPhot.table[j].sigma2FluxRatPiPj[k] = s2phot[j][k][0] / 
                    s2phot[j][k][1];
            }
        }

        /* Compute vis (diffvis) on selected frames exclusively. */
        amdlibComputeDiffVis(/* INPUT */
			     vis, wave, opd, 0, band,
			     selectedFrames,
			     (amdlibERROR_TYPE)amdlibSTATISTICAL_ERROR,
			     /* OUTPUT */
			     &imdVis, errMsg);

        /* Update complex vis (VISDATA) section accordingly. */
        amdlibAverageVis(vis, 0, band, selectedFrames, &imdVis, errMsg);

        /* In order to recompute vis2, reconvert visdata (normalized) in 
         * coherent fluxes :*/
        for (i = 0; i < nbFrames; i++)
        {
            for (j = 0; j < nbBases; j++)
            {
                for (k = 0; k < nbWlen; k++)
                {
                    P1P2 = photTabPtr[i][j].PiMultPj[k];
                    visTabPtr[i][j].vis[k].re *= (2 * amdlibSignedSqrt(P1P2));
                    visTabPtr[i][j].vis[k].im *= (2 * amdlibSignedSqrt(P1P2));
                    visTabPtr[i][j].sigma2Vis[k].re *=  (4 * P1P2);
                    visTabPtr[i][j].sigma2Vis[k].im *=  (4 * P1P2);
                }
            }
        }

        /* Compute fringe criteria on selected frames exclusively. */
        amdlibComputeFringeCriterion(vis, 0, band, selectedFrames, &imdVis);

        /* Re-Compute vis2 on selected frames exclusively */
        amdlibComputeSqVis(vis, phot, 0, band, selectedFrames, 
                           (amdlibERROR_TYPE)amdlibSTATISTICAL_ERROR,
                           &imdVis2, errMsg);
        /* update average values in header */
        amdlibAverageVis2(&imdVis2);

        /* Compute piston on selected frames exclusively */
        if (amdlibMeanPiston(opd, band, 0, selectedFrames, 
                             &imdOpd) != amdlibSUCCESS)
        {
            amdlibSetErrMsg("Could not mean piston");        
            return amdlibFAILURE;
        }

        /* Build vis3 data structure */
        if (nbBases > 1)
        {
            if (selectedFrames->band[band].nbFramesOkForClosure != 0)
            {
                /* Build vis3 values again */
                if (amdlibComputeClosurePhases(vis, 0, band, selectedFrames,
					       (amdlibERROR_TYPE)amdlibSTATISTICAL_ERROR,
					       &imdVis3,
                                               errMsg) != amdlibSUCCESS)
                {
                    return amdlibFAILURE;
                }
                amdlibAverageClosurePhases(&imdVis3);

                /* Fill VIS3 data structure */
                i = selectedFrames->band[band].frameOkForClosure[0];
                for (i = 0; i < selectedFrames->band[band].nbFramesOkForClosure; i++)
                {
                    j = selectedFrames->band[band].frameOkForClosure[i];
                    imdVis3.table[0].expTime += vis3TabPtr[j][0].expTime;
                }
            }
        }

    } 
    else
    {
        amdlibLogWarning("No frames selected for band %c", 
                         amdlibBandNumToStr(band));
    }

    /* Compute UV coords */
    for (j=0; j < nbBases; j++)
    {
        /* UV coordinates treatment */
        imdVis.table[j].uCoord = (visTabPtr[0][j].uCoord + 
                                  visTabPtr[nbFrames-1][j].uCoord) / 2;
        imdVis.table[j].vCoord = (visTabPtr[0][j].vCoord + 
                                  visTabPtr[nbFrames-1][j].vCoord) / 2;
        imdVis2.table[j].uCoord = (vis2TabPtr[0][j].uCoord + 
                                   vis2TabPtr[nbFrames-1][j].uCoord) / 2;
        imdVis2.table[j].vCoord = (vis2TabPtr[0][j].vCoord + 
                                   vis2TabPtr[nbFrames-1][j].vCoord) / 2;

        /* Time / date treatment */
        imdVis.table[j].time = (visTabPtr[0][j].time + 
                                visTabPtr[nbFrames-1][j].time) / 2;
        imdVis2.table[j].time = (vis2TabPtr[0][j].time + 
                                 vis2TabPtr[nbFrames-1][j].time) / 2;
        imdVis.table[j].dateObsMJD = (visTabPtr[0][j].dateObsMJD + 
                                      visTabPtr[nbFrames-1][j].dateObsMJD) / 2;
        imdVis2.table[j].dateObsMJD = (vis2TabPtr[0][j].dateObsMJD + 
                                       vis2TabPtr[nbFrames-1][j].dateObsMJD)/2;

        /* target identifier and station indexes treatment */
        imdVis.table[j].targetId = visTabPtr[0][j].targetId;
        imdVis.table[j].stationIndex[0] = visTabPtr[0][j].stationIndex[0];
        imdVis.table[j].stationIndex[1] = visTabPtr[0][j].stationIndex[1];

        imdVis2.table[j].targetId = vis2TabPtr[0][j].targetId;
        imdVis2.table[j].stationIndex[0] = vis2TabPtr[0][j].stationIndex[0];
        imdVis2.table[j].stationIndex[1] = vis2TabPtr[0][j].stationIndex[1];
    }
    if (nbBases > 1)
    {
        imdVis3.table[0].targetId = vis3TabPtr[0][0].targetId;
        imdVis3.table[0].time = (vis3TabPtr[0][0].time + 
                                 vis3TabPtr[nbFrames-1][0].time) / 2;
        imdVis3.table[0].dateObsMJD = (vis3TabPtr[0][0].dateObsMJD +
                                       vis3TabPtr[nbFrames-1][0].dateObsMJD)/2;
        imdVis3.table[0].stationIndex[0] = vis3TabPtr[0][0].stationIndex[0];
        imdVis3.table[0].stationIndex[1] = vis3TabPtr[0][0].stationIndex[1];
        imdVis3.table[0].stationIndex[2] = vis3TabPtr[0][0].stationIndex[2];

        imdVis3.table[0].u1Coord = (vis3TabPtr[0][0].u1Coord + 
                                    vis3TabPtr[nbFrames-1][0].u1Coord) / 2;
        imdVis3.table[0].v1Coord = (vis3TabPtr[0][0].v1Coord + 
                                    vis3TabPtr[nbFrames-1][0].v1Coord) / 2;
        imdVis3.table[0].u2Coord = (vis3TabPtr[0][0].u2Coord + 
                                    vis3TabPtr[nbFrames-1][0].u2Coord) / 2;
        imdVis3.table[0].v2Coord = (vis3TabPtr[0][0].v2Coord + 
                                    vis3TabPtr[nbFrames-1][0].v2Coord) / 2;
    }
    
    /* Free all wrappings */
    amdlibAverageVisibilities_FREEALL();

    /* Update all data structures: set computed mean values */
    free(phot->table[0].fluxSumPiPj);
    free(phot->table[0].sigma2FluxSumPiPj);
    free(phot->table[0].fluxRatPiPj);
    free(phot->table[0].sigma2FluxRatPiPj);
    free(phot->table[0].PiMultPj);
    free(phot->table);
    phot->nbFrames = 1;
    phot->table = imdPhot.table;

    free(vis->table[0].vis);
    free(vis->table[0].sigma2Vis);
    free(vis->table[0].visCovRI);
    free(vis->table[0].diffVisAmp);
    free(vis->table[0].diffVisAmpErr);
    free(vis->table[0].diffVisPhi);
    free(vis->table[0].diffVisPhiErr);
    free(vis->table[0].flag);
    free(vis->table);
    vis->nbFrames = 1;
    vis->table = imdVis.table;

    free(vis2->table[0].vis2);
    free(vis2->table[0].vis2Error);
    free(vis2->table[0].flag);
    free(vis2->table);
    vis2->nbFrames = 1;
    vis2->table = imdVis2.table;
    vis2->vis12 = imdVis2.vis12;
    vis2->vis23 = imdVis2.vis23;
    vis2->vis31 = imdVis2.vis31;
    vis2->sigmaVis12 = imdVis2.sigmaVis12;
    vis2->sigmaVis23 = imdVis2.sigmaVis23;
    vis2->sigmaVis31 = imdVis2.sigmaVis31;

    if (nbBases > 1)
    {
        free(vis3->table[0].vis3Amplitude);
        free(vis3->table[0].vis3AmplitudeError);
        free(vis3->table[0].vis3Phi);
        free(vis3->table[0].vis3PhiError);
        free(vis3->table[0].flag);
        free(vis3->table);
        vis3->nbFrames = 1;
        vis3->averageClosure = imdVis3.averageClosure;
        vis3->averageClosureError = imdVis3.averageClosureError;
        vis3->table = imdVis3.table;
    }

    opd->nbFrames = 1;
    for (i = amdlibJ_BAND; i <= amdlibK_BAND; i++)
    {
        if (opd->pistonOPDArray[i] != NULL)
        {
            free(opd->pistonOPDArray[i]);
        }
        if (opd->sigmaPistonArray[i] != NULL)
        {
            free(opd->sigmaPistonArray[i]);
        }
        opd->pistonOPDArray[i] = imdOpd.pistonOPDArray[i];
        opd->sigmaPistonArray[i] = imdOpd.sigmaPistonArray[i];
    }
    free(opd->pistonOPD);
    free(opd->sigmaPiston);
    opd->pistonOPD = imdOpd.pistonOPD;
    opd->sigmaPiston = imdOpd.sigmaPiston;

    return amdlibSUCCESS;
}

/** Useful macro to free amdlibComputeVisibilitiesFor structures */ 
#define amdlibComputeVisibilitiesFor_FREEALL()  \
    free(p2v); free(mk); free(sigma2_mk);       \
    free(v2p); free(badp2v);
/**
 * sub-entry point, gathers all strutures in input and output,
 * calls the other functions in sequence
 *
 * @param data structure containing the science observation (where the fringes   * are)
 * @param sky (optional) structure containing the SKY observation 
 * @param p2vm structure where P2VM is stored
 * @param waveData the wavelength definition structure
 * @param channelNo first channel included in computation
 * @param nbChannels number of channels total used in computation
 * @param binning the number of consecutives frames in the science observation
 * that will be used to measure a single point of visibilties in output.
 * the total time span of the frames in a bin should be of the order of
 * the coherence time of the atmosphere, unless the observation has 
 * benefitted of a fringe tracker device. At Paranal coherence time are 
 * rarely more than a few Millisecond. 
 * @param errorType indicates wether the noise figures are estimated 
 * statistically from the sequence of photometries during the bin 
 * (amdlibSTATISTICAL_ERROR) or theoretically by the poisson error on N photons
 * (amdlibTHEORETICAL_ERROR). The latter is forced obviously when the binsize
 * is 1.
 * @param pistonType indicates wether the piston is to be measured by fitting
 * a slope in the phasors amdlibITERATIVE_PHASOR or in the differential phases
 * after dewrapping (amdlibUNWRAPPED_PHASE)
 * @param noCheckP2vmId forces amdlib to use without question the passed p2vm,
 * even if its magic number is not OK. Can happen if the P2VM has been 
 * observed AFTER the science observations. 
 * @param photometry output structure containing the (binned) photometry 
 * results.
 * @param vis structure containing the (binned) raw visibilities.
 * @param vis2 structure containing the (binned) squared visibilities.
 * @param vis3 structure containing the (binned) triple products.
 * @param wave structure containing the wavelengths corresponding to the other
 * output structures. 
 * @param opd structure containing the (binned) piston measurements.
 * @param errMsg error description message returned if function fails.
 *
 * @return 
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.  */
amdlibCOMPL_STAT amdlibComputeVisibilitiesFor(/* Input */
                                              amdlibSCIENCE_DATA     *data,
                                              amdlibSCIENCE_DATA     *sky,
                                              amdlibP2VM_MATRIX      *p2vm,
                                              amdlibWAVEDATA         *waveData,
                                              int                    *channelNo,
                                              int                    nbChannels,
                                              int                    binning,
                                              amdlibERROR_TYPE       errorType,
                                              amdlibPISTON_ALGORITHM pistonType,
                                              amdlibBOOLEAN       noCheckP2vmId,
                                              /* Output */
                                              amdlibPHOTOMETRY      *photometry,
                                              amdlibVIS              *vis,
                                              amdlibVIS2             *vis2,
                                              amdlibVIS3             *vis3,
                                              amdlibWAVELENGTH       *wave,
                                              amdlibPISTON           *opd,
                                              amdlibERROR_MSG        errMsg)
{

    int nbPix = data->col[amdlibINTERF_CHANNEL].nbPixels;
    int nbTel = data->nbCols - 1;  /* Number of telescopes used in data*/
    int nbFrames = data->nbFrames;
    int nbBases = nbTel * (nbTel - 1) / 2; /* Number of baselines in data*/
    int nbClos = nbTel * (nbTel - 1) * (nbTel - 2) / 6; /* Number of closures */
    int iBin; /* Index of the current binning */
    int firstFrame; /* Number of visibilities stored in OI_VIS */
    int nbBins; /* Number of binnings */
    double *p2v; /* local copy of (eventual 2T subset of) current p2vm */
    unsigned char *badp2v; /* local copy of current p2vm Bad Pixels */
    double *v2p; /* the direct pixel to visibility matrix */
    double *mk;
    double *sigma2_mk;
    static amdlibVIS instantCorrFlux;
    static amdlibPHOTOMETRY instantPhotometry;
    static amdlibPISTON instantOpd;
    amdlibBAND band;
    amdlibSELECTION selection;

    amdlibLogTrace("amdlibComputeVisibilitiesFor()");

    if(binning > nbFrames)
    {
        binning = nbFrames;
        nbBins = 1;
    }
    else
    {
        nbBins = data->nbFrames / binning;
    }

    if (binning == 1)
    {
        errorType = amdlibTHEORETICAL_ERROR; 
    }
#if amdlibDEBUG
    amdlibLogTrace("amdlibComputeVisibilitiesFor() - first data : ");
    amdlibLogTrace("nbPix = %d", nbPix);
    amdlibLogTrace("nbTel = %d", nbTel);
    amdlibLogTrace("nbFrames = %d", nbFrames);
    amdlibLogTrace("nbBases = %d", nbBases);
    amdlibLogTrace("nbClos = %d", nbClos);
    amdlibLogTrace("nbBins = %d", nbBins);    
    amdlibLogTrace("nbChannels = %d", nbChannels);
    amdlibLogTrace("details: ");
    int i;
    for (i = 0; i < nbChannels; i++)
    {
        amdlibLogTrace("channelNo[%d] = %d, wlen = %f", i, channelNo[i], 
                       waveData->wlen[channelNo[i]]);
    }
#endif

    if(amdlibCheckInsConfig(data, sky, p2vm, noCheckP2vmId, 
                            errMsg) == amdlibFAILURE)
    {
        return amdlibFAILURE;
    }

    /* Allocate all the structures used by amdlibComputeVisibilitiesFor */
    /* Allocate Wavelength Structure */
    if (amdlibAllocateWavelength(wave, nbChannels, errMsg) != amdlibSUCCESS)
    {
        return amdlibFAILURE;
    }
    /* Allocate memory for photometry */
    if (amdlibAllocatePhotometry(photometry, nbBins, nbBases, 
                                 nbChannels) != amdlibSUCCESS)
    {
        amdlibSetErrMsg("Could not allocate memory for PHOTOMETRY");
        return amdlibFAILURE;
    }
    /* Allocate memory for vis */
    if (amdlibAllocateVis(vis, nbBins, nbBases, nbChannels) != amdlibSUCCESS)
    {
        amdlibSetErrMsg("Could not allocate memory for VIS");
        return amdlibFAILURE;
    }
    /* Allocate memory vis2  */
    if (amdlibAllocateVis2(vis2, nbBins, nbBases, nbChannels) != amdlibSUCCESS)
    {
        amdlibSetErrMsg("Could not allocate memory for VIS2");
        return amdlibFAILURE;
    }
    if (nbTel == 3)
    {
        /* Allocate VIS3 structure */
        if (amdlibAllocateVis3(vis3, nbBins, nbClos, 
                               nbChannels) != amdlibSUCCESS)
        {
            amdlibSetErrMsg("Could not allocate memory for VIS3");
            return amdlibFAILURE;
        }
    }
    /* Allocate Piston Structure */
    if (amdlibAllocatePiston(opd, nbBins, nbBases) != amdlibSUCCESS)
    {
        amdlibSetErrMsg("Could not allocate memory for Piston structure");
        return amdlibFAILURE;
    }

    /* Allocate p2v matrix and associated matrixes */
    /* local copy of (eventual 2T subset of) current
     *  p2vm dim(nbPix,2*nbBases)
     */
    p2v       = calloc(2*nbBases*nbPix*nbChannels, sizeof(double)); 
    /* local copy of (eventual 2T subset of)
     * current p2vm Bad Pixels 
     */
    badp2v    = calloc(nbChannels*nbPix, sizeof(char)); 
    mk        = calloc(binning*nbChannels*nbPix, sizeof(double)); 
    sigma2_mk = calloc(binning*nbChannels*nbPix, sizeof(double));
    /* the direct visibility to pixel matrix */
    v2p       = calloc(nbPix*2*nbBases*nbChannels, sizeof(double));
    /* Allocate memory for the instantaneous detector noise time series*/

    /* Allocate memory for the instantaneous photometry
     * used for all the computations */
    if (amdlibAllocatePhotometry(&instantPhotometry,
                                 binning, nbBases, nbChannels) != amdlibSUCCESS)
    {
        amdlibSetErrMsg("Could not allocate memory for instantPhotometry");
        amdlibComputeVisibilitiesFor_FREEALL();
        return amdlibFAILURE;
    }

    /* Allocate memory for the instantaneous visibility
     * used for all the computations */
    if (amdlibAllocateVis(&instantCorrFlux,
                          binning, nbBases, nbChannels) !=
        amdlibSUCCESS)
    {
        amdlibSetErrMsg("Could not allocate memory for instantCorrFlux");
        amdlibReleasePhotometry(&instantPhotometry);
        amdlibComputeVisibilitiesFor_FREEALL();
        return amdlibFAILURE;
    }
    /* Allocate memory for the instantaneous piston time series*/
    if (amdlibAllocatePiston(&instantOpd, binning, nbBases) !=
        amdlibSUCCESS)
    {
        amdlibSetErrMsg("Could not allocate memory for instantOpd");
        amdlibComputeVisibilitiesFor_FREEALL();
        amdlibReleasePhotometry(&instantPhotometry);
        amdlibReleaseVis(&instantCorrFlux);
        return amdlibFAILURE;
    }

    if (amdlibAllocateSelection(&selection, 
                                binning, nbBases, errMsg) != amdlibSUCCESS)
    {
        amdlibComputeVisibilitiesFor_FREEALL();
        amdlibReleasePhotometry(&instantPhotometry);
        amdlibReleaseVis(&instantCorrFlux);
        return amdlibFAILURE;
    }

    for(iBin=0; iBin < nbBins; iBin++)
    {
        /* The first frame to read in the data */
        firstFrame = iBin * binning;
        /*amdlibLogTrace("binning = %d", binning); */


        /* Get the right data, using the P2VM and data rows informations */
        if (amdlibCrossP2vmAndData(/* Input */
                                   p2vm, data, sky, waveData, firstFrame,
                                   binning, channelNo, nbChannels,
                                   /* Output */
                                   wave,
                                   &instantPhotometry,
                                   p2v, badp2v,
                                   mk, sigma2_mk,
                                   errMsg) != amdlibSUCCESS)
        {
            amdlibComputeVisibilitiesFor_FREEALL();
            amdlibReleasePhotometry(&instantPhotometry);
            amdlibReleaseVis(&instantCorrFlux);
            amdlibReleasePiston(&instantOpd);
            return amdlibFAILURE;
        }

        /* Invert the P2VM, using a generalized inverse */
        if (amdlibInvertP2v(/* Input */
                            p2v, badp2v, sigma2_mk, 
                            binning, nbChannels, nbPix, nbBases,
                            /* Output */
                            v2p, errMsg) != amdlibSUCCESS)
        {
            amdlibComputeVisibilitiesFor_FREEALL();
            amdlibReleasePhotometry(&instantPhotometry);
            amdlibReleaseVis(&instantCorrFlux);
            amdlibReleasePiston(&instantOpd);
            return amdlibFAILURE;
        }

        /* Compute correlated flux, used after to compute visibilities */
        if (amdlibComputeCorrFlux(/* Input */
                                  v2p, mk, sigma2_mk,
                                  binning, nbChannels, nbPix, nbBases,
                                  /* Output */
                                  &instantCorrFlux, errMsg) != amdlibSUCCESS)
        {
            amdlibComputeVisibilitiesFor_FREEALL();
            amdlibReleasePhotometry(&instantPhotometry);
            amdlibReleaseVis(&instantCorrFlux);
            amdlibReleasePiston(&instantOpd);
            return amdlibFAILURE;
        }


        /* Achromatic piston computation */
        band = amdlibGetBand(wave->wlen[0]);

        if (amdlibComputePiston(&instantCorrFlux, band, wave, pistonType, 
                                &instantOpd, errMsg) != amdlibSUCCESS)
        {
            amdlibComputeVisibilitiesFor_FREEALL();
            amdlibReleasePhotometry(&instantPhotometry);
            amdlibReleaseVis(&instantCorrFlux);
            return amdlibFAILURE;
        }

        if (amdlibMeanPiston(/* Input */
                             &instantOpd, band, iBin, &selection,
                             /* Output */
                             opd) != amdlibSUCCESS)
        {
            amdlibComputeVisibilitiesFor_FREEALL();
            amdlibReleasePhotometry(&instantPhotometry);
            amdlibReleaseVis(&instantCorrFlux);
            amdlibReleasePiston(&instantOpd);
            return amdlibFAILURE;
        }


        /* Fringe criterion (SNR) computation */
        amdlibComputeFringeCriterion(/* Input */
                                     &instantCorrFlux, iBin, band, &selection,
                                     /* Output */
                                     vis);

        /* complex visibilities average */
        if (amdlibMeanCpxVis(/* Input */
                             &instantCorrFlux, 
                             &instantPhotometry,
                             iBin,
                             /* Output */
                             vis) != amdlibSUCCESS)
        {
            amdlibComputeVisibilitiesFor_FREEALL();
            amdlibReleasePhotometry(&instantPhotometry);
            amdlibReleaseVis(&instantCorrFlux);
            amdlibReleasePiston(&instantOpd);
            return amdlibFAILURE;
        }

        /* squared visibilities estimation */
        if (amdlibComputeSqVis(/* Input */
                               &instantCorrFlux, &instantPhotometry,
                               iBin, band, &selection, errorType,
                               /* Output */
                               vis2, errMsg) != amdlibSUCCESS)
        {
            amdlibComputeVisibilitiesFor_FREEALL();
            amdlibReleasePhotometry(&instantPhotometry);
            amdlibReleaseVis(&instantCorrFlux);
            amdlibReleasePiston(&instantOpd);
            return amdlibFAILURE;
        }

        /* flux computation */
        if (amdlibSumPhotometry(/* Input */
				&instantPhotometry,
                                iBin,
				/* Output */
                                photometry) != amdlibSUCCESS)
        {
            amdlibComputeVisibilitiesFor_FREEALL();
            amdlibReleasePhotometry(&instantPhotometry);
            amdlibReleaseVis(&instantCorrFlux);
            amdlibReleasePiston(&instantOpd);
            return amdlibFAILURE;
        }


        /* Compute Closure phases. */
        if (nbTel == 3)
        {
            if (amdlibComputeClosurePhases(/* Input */
					   &instantCorrFlux, iBin, band,
					   &selection, errorType,
					   /* Output */
					   vis3, errMsg) != amdlibSUCCESS)
            {
                amdlibComputeVisibilitiesFor_FREEALL();
                amdlibReleasePhotometry(&instantPhotometry);
                amdlibReleaseVis(&instantCorrFlux);
                amdlibReleasePiston(&instantOpd);
                return amdlibFAILURE;
            }
        } /*End if (nbTel == 3) */

        /* Compute diffvis */
        if (amdlibComputeDiffVis(/* Input */
				 &instantCorrFlux, wave, &instantOpd, 
                                 iBin, band, &selection, errorType,
				 /* Output */
				 vis, errMsg)!= amdlibSUCCESS)
        {
            amdlibComputeVisibilitiesFor_FREEALL();
            amdlibReleasePhotometry(&instantPhotometry);
            amdlibReleaseVis(&instantCorrFlux);
            amdlibReleasePiston(&instantOpd);
            return amdlibFAILURE;
        }
    }

    /* Finalize table headers */
    amdlibFillInVisTableHeader(data, vis, errMsg);
    amdlibFillInVis2TableHeader(data, vis2, errMsg);
    if (nbTel == 3) 
    {
        amdlibFillInVis3TableHeader(data, vis3, errMsg);
    }

    amdlibComputeVisibilitiesFor_FREEALL();
    amdlibReleasePhotometry(&instantPhotometry);
    amdlibReleaseVis(&instantCorrFlux);
    amdlibReleasePiston(&instantOpd);
    amdlibAverageVis2(vis2);

    amdlibAverageClosurePhases(vis3);

    amdlibReleaseSelection(&selection);

    return amdlibSUCCESS;
}
#undef amdlibComputeVisibilitiesFor_FREEALL

/*
 * Local functions definitions
 */
/**
 * Check instrument configuration and coherence of the files used in the data
 * reduction.
 *
 * @param data structure containing the science observation (where the fringes 
 * are)
 * @param sky (optional) structure containing the SKY observation 
 * @param p2vm structure where P2VM is stored
 * @param noCheckP2vmId forces amdlib to use without question the passed p2vm,
 * @param errMsg error description message returned if function fails.
 *
 * @return 
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.  */
amdlibCOMPL_STAT amdlibCheckInsConfig(/* Input */
                                      amdlibSCIENCE_DATA     *data,
                                      amdlibSCIENCE_DATA     *sky,
                                      amdlibP2VM_MATRIX      *p2vm,
                                      amdlibBOOLEAN          noCheckP2vmId,
                                      /* Output */
                                      amdlibERROR_MSG        errMsg)
{

    int nbTelInp2v = amdlibNB_TEL; 
    /* Number of Telescopes Used in the p2vm calibration file*/
    int nbTel = data->nbCols - 1; 
    /* Number of Telescopes Used in the Science data*/

    amdlibLogTrace("amdlibCheckInsConfig()");

    if (noCheckP2vmId == amdlibFALSE)
    {
        /* Check data reference to P2VM id */ 
        if (data->p2vmId == -1)
        {
            amdlibSetErrMsg("No P2VM available for this data");
            return amdlibFAILURE;
        }
        /* Check P2VM Ids */
        if ((p2vm->id != 0) && (data->p2vmId != 0) && 
            (p2vm->id != data->p2vmId))
        {
            amdlibSetErrMsg("PV2M id '%d' referenced in data does not match "
                            "with PV2M id '%d'", data->p2vmId, p2vm->id);
            return amdlibFAILURE;
        }
        /* If sky is specified, check it too */
        if ((sky != NULL) && (noCheckP2vmId == amdlibFALSE))
        {
            if ((p2vm->id != 0) && (sky->p2vmId != 0) && 
                (p2vm->id != sky->p2vmId))
            {
                amdlibSetErrMsg("PV2M id '%d' referenced in sky does not match "
                                "with PV2M id '%d'", sky->p2vmId, p2vm->id);
                return amdlibFAILURE;
            }
        }
    }

    /* If P2VM id was not written in header */
    if ((p2vm->id == 0) || (data->p2vmId == 0))
    {
        /* Check instrument configuration */
        int keywIdx = 0;
        int dataKeywIdx;
        int skyKeywIdx;
        int p2vmKeywIdx;
        while (amdlibP2vmInsCfgKeywList[keywIdx] != NULL)
        {
            /* Found keyword in data instrument cfg */
            dataKeywIdx = 0;
            while ((strlen(data->insCfg.keywords[dataKeywIdx].name) != 0) &&
                   (strncmp(data->insCfg.keywords[dataKeywIdx].name, 
                            amdlibP2vmInsCfgKeywList[keywIdx],
                            strlen(amdlibP2vmInsCfgKeywList[keywIdx])) != 0))
            {
                dataKeywIdx++;
            }

            /* Found keyword in sky instrument cfg */
            skyKeywIdx = 0;
            if (sky != NULL)
            {
                while ((strlen(sky->insCfg.keywords[skyKeywIdx].name) != 0) &&
                       (strncmp(sky->insCfg.keywords[skyKeywIdx].name, 
                            amdlibP2vmInsCfgKeywList[keywIdx],
                            strlen(amdlibP2vmInsCfgKeywList[keywIdx])) != 0))
                {
                    skyKeywIdx++;
                }
            }

            /* Found keyword in P2VM instrument cfg */
            p2vmKeywIdx = 0;
            while ((strlen(p2vm->insCfg.keywords[p2vmKeywIdx].name) != 0) &&
                   (strncmp(p2vm->insCfg.keywords[p2vmKeywIdx].name, 
                            amdlibP2vmInsCfgKeywList[keywIdx],
                            strlen(amdlibP2vmInsCfgKeywList[keywIdx])) != 0))
            {
                p2vmKeywIdx++;
            }


            /* Check keyword values (if found) */
            if ((strlen(data->insCfg.keywords[dataKeywIdx].name) != 0) &&
                (strlen(p2vm->insCfg.keywords[p2vmKeywIdx].name) != 0))
            {
                if (strcmp(data->insCfg.keywords[dataKeywIdx].value,
                           p2vm->insCfg.keywords[p2vmKeywIdx].value) != 0)
                {
                    amdlibStripBlanks(data->insCfg.keywords[dataKeywIdx].value);
                    amdlibStripBlanks(p2vm->insCfg.keywords[p2vmKeywIdx].value);

                    amdlibSetErrMsg("Value of keyword'%s' differs for P2VM "
                                    "'%s' and data '%s'",
                                    amdlibP2vmInsCfgKeywList[keywIdx],
                                    p2vm->insCfg.keywords[p2vmKeywIdx].value,
                                    data->insCfg.keywords[dataKeywIdx].value);
                    return amdlibFAILURE;
                }
            }

            if (sky != NULL)
            {
                if ((strlen(sky->insCfg.keywords[skyKeywIdx].name) != 0) &&
                    (strlen(p2vm->insCfg.keywords[p2vmKeywIdx].name) != 0))
                {
                    if (strcmp(sky->insCfg.keywords[skyKeywIdx].value,
                               p2vm->insCfg.keywords[p2vmKeywIdx].value) != 0)
                    {
                        amdlibStripBlanks(
                                    sky->insCfg.keywords[skyKeywIdx].value);
                        amdlibStripBlanks(
                                    p2vm->insCfg.keywords[p2vmKeywIdx].value);

                        amdlibSetErrMsg("Value of keyword '%s' differs for "
                                        "P2VM '%s' and sky '%s'",
                                        amdlibP2vmInsCfgKeywList[keywIdx],
                                        p2vm->insCfg.keywords[p2vmKeywIdx].value,
                                        sky->insCfg.keywords[skyKeywIdx].value);
                        return amdlibFAILURE;
                    }
                }
            }

            /* Check next keyword */
            keywIdx++;
        }
    }

    /* Init the number of telescopes and baselines according to P2VM type */
    if (p2vm->type == amdlibP2VM_2T)
    {
        nbTelInp2v =  2;
    }
    else if (p2vm->type == amdlibP2VM_3T)
    {
        nbTelInp2v =  3;
    }
    else
    {
        amdlibSetErrMsg("Invalid P2VM type %d (see amdlibP2VM_TYPE)",
                        (int)p2vm->type);
        return amdlibFAILURE;
    }
    /* Should Check the number of telescopes and baselines according to
     * what the Science data claims
     */
    nbTel = data->nbCols - 1;
    if (nbTel > nbTelInp2v)
    {
        amdlibSetErrMsg("P2VM is %dT and Data %dT, aborting.",
                        nbTelInp2v, nbTel);
        return amdlibFAILURE;
    }
    /*
     * Make a few checks on the adequation between p2vm and science data
     */
    if (p2vm->nx != data->col[amdlibINTERF_CHANNEL].nbPixels)
    {
        amdlibSetErrMsg("Incompatible P2VM and data interf channel width "
                        "(%d vs. %d)", p2vm->nx,
                        data->col[amdlibINTERF_CHANNEL].nbPixels);
        return amdlibFAILURE;
    }

    if (data->nbChannels <= 0)
    {
        amdlibSetErrMsg("Invalid nbWlen (%d) information in data.",
                        data->nbChannels);
        return amdlibFAILURE;
    }
    /* I do not understand the following test. Corrected to accept (at least) 
     * simulated data from amdsim whose 1st valid spectral channel is 0
     * . GD */
    if(data->channelNo[0] < 0)
    {
        amdlibSetErrMsg("Invalid startPixel (%d) information in data.", 
                        data->channelNo[0]);
        return amdlibFAILURE;
    }
    return amdlibSUCCESS;

}


/**
 * Useful macro to free memory allocted in amdlibCrossP2vmAndData function
 */
#define amdlibCrossP2vmAndData_FREEALL()                                \
    amdlibFree2DArrayDouble(vk);                                        \
    amdlibFree3DArrayDouble(phot);                                      \
    amdlibFree3DArrayDouble(sigma2Phot);                                \
    amdlibFree3DArrayDoubleWrapping(p2vCkDkPtr);                        \
    amdlibFree3DArrayDoubleWrapping(mkPtr);                             \
    amdlibFree3DArrayDoubleWrapping(sigma2MkPtr);                       \
    amdlibFree2DArrayDouble(sumVk);                                     \
    amdlibFree2DArrayUnsignedCharWrapping(badPixelFlagsPtr);            \
    amdlibFree2DArrayWrapping((void **)tablePt);                        \
    free(flatValues);
/**
 * Extracts in the (large) input structures the relevant subset of values
 * needed, takes care of substracting the sky data, computes the mk and 
 * photometries.
 *
 * @param p2vm pointer to P2VM.
 * @param data pointer to science data.
 * @param sky pointer to sky data.
 * @param waveData the wavelength definition structure
 * @param firstFrame first frame in science data to handle.
 * @param nbFrames number of frames in science data to handle.
 * @param channelNo indexes of valid spectral channels to treat.
 * @param nbChannels number of valid spectral channels to treat.
 * @param wave structure containing the wavelengths corresponding to the other
 * @param photometry output structure containing the (binned) photometry 
 * results.
 * @param p2vCkDk ck and dk values of P2VM.
 * @param badPixelFlags pointer to the bad pixel map
 * @param mk the continumm-corrected values of the interferograms
 * @param sigma2Mk the associated variances of the mks
 * @param errMsg error description message returned if function fails.
 *
 * @return 
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.  */
amdlibCOMPL_STAT amdlibCrossP2vmAndData(/* Input */
                                        amdlibP2VM_MATRIX  *p2vm,
                                        amdlibSCIENCE_DATA *data,
                                        amdlibSCIENCE_DATA *sky,
                                        amdlibWAVEDATA     *waveData,
                                        int                firstFrame,
                                        int                nbFrames,
                                        int                *channelNo,
                                        int                nbChannels,
                                        /* Output */
                                        amdlibWAVELENGTH   *wave,
                                        amdlibPHOTOMETRY   *photometry,
                                        double             *p2vCkDk,
                                        unsigned char      *badPixelFlags,
                                        double             *mk,
                                        double             *sigma2Mk,
                                        amdlibERROR_MSG    errMsg)
{

    /* loop indexes and associated limits */
    /* Index on pixels */
    int iPix;
    int nbPix = data->col[amdlibINTERF_CHANNEL].nbPixels;

    /* Index on telescopse */
    int iTel;
    int nbTel = data->nbCols - 1;

    /* Index on baselines */
    int iBase;
    int nbBases = nbTel * (nbTel - 1) / 2;

    /* Index on frames */
    int iFrame;

    /* Index on spectral channel in P2VM */
    int iP2vmChannel;

    /* Index on spectral channel in scienData */
    int iDataChannel;

    /* Index on spectral channel */
    int channel;

    /* Array to store Vk and sumVk */
    double **vk = NULL;
    double **sumVk = NULL;
    double ***phot = NULL;
    double ***sigma2Phot = NULL;
    double ***p2vCkDkPtr = NULL;
    double ***mkPtr = NULL;
    double ***sigma2MkPtr = NULL;
    float *flatValues = NULL;
    unsigned char **badPixelFlagsPtr = NULL;

    int tel1, tel2;
    double P1P2, P1, P2, sigma2P1, sigma2P2;


    amdlibPHOTOMETRY_TABLE_ENTRY **tablePt = NULL;

    amdlibLogTrace("amdlibCrossP2vmAndData()"); 

    /* Flag indicating whether sky is given or not */
    amdlibBOOLEAN hasSky;
    hasSky=(sky==NULL)?amdlibFALSE:amdlibTRUE;

    /* Check that science data contains the expected frames */
    if(firstFrame + nbFrames > data->nbFrames)
    {
        amdlibSetErrMsg("Specified number of frames overreaches the total "
                        "number of frames.");
        return amdlibFAILURE;
    }

    /* If specified, check that sky data only contains one frame */
    if (hasSky == amdlibTRUE)
    {
        if (sky->nbFrames > 1)
        {
            amdlibSetErrMsg("Invalid Multi-Frame Sky found.");
            return amdlibFAILURE;
        }
    }

    /* Allocate memory */
    vk = amdlibAlloc2DArrayDouble(nbPix, nbTel, errMsg);
    if (vk == NULL)
    {
        amdlibCrossP2vmAndData_FREEALL();
        return amdlibFAILURE;
    }
    sumVk = amdlibAlloc2DArrayDouble(nbBases, nbChannels, errMsg);
    if (sumVk == NULL)
    {
        amdlibCrossP2vmAndData_FREEALL();
        return amdlibFAILURE;
    }
 
    phot = amdlibAlloc3DArrayDouble(nbTel, nbChannels, nbFrames, errMsg);
    if (phot == NULL)
    {
        amdlibCrossP2vmAndData_FREEALL();
        return amdlibFAILURE;
    }
    sigma2Phot = amdlibAlloc3DArrayDouble(nbTel, nbChannels, nbFrames, errMsg);
    if (sigma2Phot == NULL)
    {
        amdlibCrossP2vmAndData_FREEALL();
        return amdlibFAILURE;
    }
    flatValues = calloc(nbPix, sizeof(float));
    if (flatValues == NULL)
    {
        amdlibCrossP2vmAndData_FREEALL();
        return amdlibFAILURE;
    }

    /* And wrap already allocated array */
    p2vCkDkPtr = amdlibWrap3DArrayDouble(p2vCkDk, 2 * nbBases, nbPix,
                                         nbChannels, errMsg);
    if (p2vCkDkPtr == NULL)
    {
        amdlibCrossP2vmAndData_FREEALL();
        return amdlibFAILURE;
    }
    mkPtr = amdlibWrap3DArrayDouble(mk, nbPix, nbChannels, nbFrames, errMsg);
    if (mkPtr == NULL)
    {
        amdlibCrossP2vmAndData_FREEALL();
        return amdlibFAILURE;
    }
    sigma2MkPtr = amdlibWrap3DArrayDouble(sigma2Mk, nbPix, nbChannels, nbFrames,
                                          errMsg);
    if (sigma2MkPtr == NULL)
    {
        amdlibCrossP2vmAndData_FREEALL();
        return amdlibFAILURE;
    }
    badPixelFlagsPtr = amdlibWrap2DArrayUnsignedChar(badPixelFlags, nbPix, 
                                                     nbChannels, errMsg);
    if (badPixelFlagsPtr == NULL)
    {
        amdlibCrossP2vmAndData_FREEALL();
        return amdlibFAILURE;
    }

    /* For each channel */
    tablePt = (amdlibPHOTOMETRY_TABLE_ENTRY **)
        amdlibWrap2DArray(photometry->table,  nbBases, nbFrames, 
                          sizeof(*(photometry->table)), errMsg);

    for (channel = 0; channel < nbChannels; channel++)
    {
        /* Check this channel is valid in P2VM, and get entry in P2VM
         * corresponding to channel */
        if (amdlibIsValidChannel(p2vm, channelNo[channel],
                                 &iP2vmChannel) == amdlibFALSE)
        {
            amdlibSetErrMsg("Channel '%d' is not in P2VM or it is not valid",
                            channelNo[channel]);
            return amdlibFAILURE;
        }

        /* Check this channel belows to science data */
        iDataChannel = 0;
        while ((data->channelNo[iDataChannel] != channelNo[channel]) &&
               (iDataChannel < data->nbChannels))
        {
            iDataChannel++;
        }
        if (iDataChannel == data->nbChannels)
        {
            amdlibSetErrMsg("Channel '%d' does not below to science data", 
                            channelNo[channel]);
            return amdlibFAILURE;
        }

#if amdlibDEBUG
        amdlibLogTrace("channelNo=%d, iDataChannel=%d, iP2vmChannel=%d, "
                       "wlen=%f",
                       channelNo[channel], iDataChannel, iP2vmChannel, 
               p2vm->wlen[iP2vmChannel]);
#endif

        /* Extract cks, dks for all pixel of spectral channel from P2VM and
         * store into p2v */
        for (iBase = 0; iBase < (2 * nbBases); iBase++)
        {
            for (iPix = 0; iPix < nbPix; iPix++)
            {
                p2vCkDkPtr[channel][iPix][iBase] = 
                    p2vm->matrixPt[iP2vmChannel][iPix][iBase];
#if amdlibDEBUG
                amdlibLogTrace("p2vCkDkPtr[%d][%d][%d] = %f", channel, iPix, 
                               iBase, p2vCkDkPtr[channel][iPix][iBase]); 
#endif
            }
        }
        /* Then extract bad pixels and flat-field values */
        for (iPix = 0; iPix < nbPix; iPix++)
        {
            /* Bad P2v pixels*/
            badPixelFlagsPtr[channel][iPix] =
                p2vm->badPixelsPt[iP2vmChannel][iPix];
            flatValues[iPix]   =
                p2vm->flatFieldPt[iP2vmChannel][iPix];
        }

        /* Then extract Vk */
        for (iTel = 0; iTel < nbTel; iTel++)
        {
            for (iPix = 0; iPix < nbPix; iPix++)
            { /* begin loop for iPix */
                vk[iTel][iPix] = p2vm->vkPt[iTel][iP2vmChannel][iPix];
            }
        }

        /* Then extract sumVk */
        for (iBase = 0; iBase < nbBases; iBase++)
        {
            sumVk[channel][iBase] = p2vm->sumVkPt[iBase][iP2vmChannel];
        }

        /* Extract time sequence of photometry and associated noise, and
         * normalize them */
        for (iFrame = 0; iFrame < nbFrames; iFrame++)
        {
            float *photData[amdlibNB_TEL];
            float *sigma2PhotData[amdlibNB_TEL];
            float *photSky[amdlibNB_TEL];
            float *sigma2PhotSky[amdlibNB_TEL];
            photData[0] = data->frame[iFrame + firstFrame].photo1;
            photData[1] = data->frame[iFrame + firstFrame].photo2;
            photData[2] = data->frame[iFrame + firstFrame].photo3;
            sigma2PhotData[0] = data->frame[iFrame + firstFrame].sigma2Photo1;
            sigma2PhotData[1] = data->frame[iFrame + firstFrame].sigma2Photo2;
            sigma2PhotData[2] = data->frame[iFrame + firstFrame].sigma2Photo3;
            if (hasSky == amdlibTRUE)
            {
                photSky[0] = sky->frame[0].photo1;
                photSky[1] = sky->frame[0].photo2;
                photSky[2] = sky->frame[0].photo3;
                sigma2PhotSky[0] = sky->frame[0].sigma2Photo1;
                sigma2PhotSky[1] = sky->frame[0].sigma2Photo2;
                sigma2PhotSky[2] = sky->frame[0].sigma2Photo3;
            }
            /* extract time sequence of Phot and associated sigma2Phot */
            if (hasSky == amdlibTRUE)
            {
                for (iTel = 0; iTel < nbTel; iTel++) 
                {
                    phot[iFrame][channel][iTel] = photData[iTel][iDataChannel] -
                        (data->expTime / sky->expTime) * 
                        photSky[iTel][iDataChannel];
#if amdlibDEBUG
                    amdlibLogTrace("phot[%d][%d][%d] = %f", iFrame, channel, 
                                   iTel, phot[iFrame][channel][iTel]); 
#endif

                    sigma2Phot[iFrame][channel][iTel] = 
                        sigma2PhotData[iTel][iDataChannel] +
                        (data->expTime / sky->expTime) *
                        sigma2PhotSky[iTel][iDataChannel];
                }
            }
            else
            {
                for (iTel = 0; iTel < nbTel; iTel++) 
                {
                    phot[iFrame][channel][iTel] = photData[iTel][iDataChannel] ;
                    sigma2Phot[iFrame][channel][iTel] =
                        sigma2PhotData[iTel][iDataChannel];
                }
            }
        }

        /* Calculation of the spectrum informations */
        for (iFrame = 0; iFrame < nbFrames; iFrame++)
        {
            for (iBase = 0; iBase < nbBases; iBase++)
            { 
                /* Set used telescope according to the baseline */
                if (iBase == amdlibBASE12)
                {
                    tel1 = amdlibTEL1;
                    tel2 = amdlibTEL2;
                }
                else if (iBase == amdlibBASE23)
                {
                    tel1 = amdlibTEL2;
                    tel2 = amdlibTEL3;
                }
                else
                {
                    tel1 = amdlibTEL1;
                    tel2 = amdlibTEL3;
                }
                /* P1P2 is directly estimated, giving PiMultPj */
                P1P2 = phot[iFrame][channel][tel1] * 
                    phot[iFrame][channel][tel2];

                tablePt[iFrame][iBase].PiMultPj[channel] = 
                P1P2 * sumVk[channel][iBase];

                /* Get integrated photometry and associated noise for each 
                 * beam . Involves a summation on interferometry pixels.*/
                P1 = 0.0;
                P2 = 0.0;
                sigma2P1 = 0.0;
                sigma2P2 = 0.0;
                for (iPix = 0; iPix < nbPix; iPix++)
                { 
                    P1 += phot[iFrame][channel][tel1] * vk[tel1][iPix];
                    sigma2P1 += 
                        sigma2Phot[iFrame][channel][tel1] * vk[tel1][iPix];

                    P2 += phot[iFrame][channel][tel2] * vk[tel2][iPix];
                    sigma2P2 += 
                        sigma2Phot[iFrame][channel][tel2] * vk[tel2][iPix];
                }
                
                /* Get integrated photometry and associated noise for both.
                 * Since we want to be able afterwards to retrieve the 
                 * individual photometries and errors, the errors of ratio
                 * is just ratio of errors, (its just a coding) */
                tablePt[iFrame][iBase].fluxSumPiPj[channel] = P1 + P2;

                tablePt[iFrame][iBase].sigma2FluxSumPiPj[channel] =
                sigma2P1 + sigma2P2;          

                tablePt[iFrame][iBase].fluxRatPiPj[channel] = 
                P1/P2;

                tablePt[iFrame][iBase].sigma2FluxRatPiPj[channel] =
                sigma2P1/sigma2P2;
            }
        }

        /* Calculation of the continuum corrected interferogram.
         * See eq 24 */
        for (iFrame = 0; iFrame < nbFrames; iFrame++)
        {
            double  sigmaDet2;
            double  sigmaDet2Sky=0.0;
            float **intfPtr = NULL;
            float **intfSkyPtr = NULL;

            sigmaDet2 = amdlibPow2(data->frame[iFrame + firstFrame].ron);
            if (hasSky == amdlibTRUE)
            {
                sigmaDet2Sky = amdlibPow2(sky->frame[0].ron);
            }

            intfPtr = 
                amdlibWrap2DArrayFloat(data->frame[iFrame+firstFrame].intf,
                                       nbPix, data->nbChannels, errMsg);
            if (intfPtr == NULL)
            {
                amdlibCrossP2vmAndData_FREEALL();
                return amdlibFAILURE;
            }

            /* Wrap interferometric channels */
            if (hasSky == amdlibTRUE)
            {
                intfSkyPtr = amdlibWrap2DArrayFloat(sky->frame[0].intf,
                                            nbPix, data->nbChannels, errMsg);
                if (intfSkyPtr == NULL)
                {
                    amdlibCrossP2vmAndData_FREEALL();
                    return amdlibFAILURE;
                }
            }

            for (iPix = 0; iPix < nbPix; iPix++)
            {
                /* Extract time sequence of mk = interf - sum_tel(vk*Phot) 
                 * and associated sigma2_mk */
                /* Compute sigma2_mk values using the fact that mk pixels 
                 * values are deflated e-, (mk=e-/flat), whereas the 
                 * variance of these pixels is sigma=sqrt(e-)/flat
                 * thus sigma2=e-/flat^2 = mk/flat
                 */
                if (hasSky == amdlibTRUE)
                {
                    mkPtr[iFrame][channel][iPix] =
                        intfPtr[iDataChannel][iPix] - 
                        (data->expTime / sky->expTime) * 
                        intfSkyPtr[iDataChannel][iPix];
                    sigma2MkPtr[iFrame][channel][iPix] =
                        intfPtr[iDataChannel][iPix] / flatValues[iPix] +
                        sigmaDet2 / amdlibPow2(flatValues[iPix]) +
                        (data->expTime / sky->expTime) * 
                        (intfSkyPtr[iDataChannel][iPix] / flatValues[iPix] +
                         sigmaDet2Sky / amdlibPow2(flatValues[iPix]));
                }
                else
                {
                    mkPtr[iFrame][channel][iPix] = intfPtr[iDataChannel][iPix];
                    sigma2MkPtr[iFrame][channel][iPix] =
                        intfPtr[iDataChannel][iPix] / flatValues[iPix] + 
                        sigmaDet2 / amdlibPow2(flatValues[iPix]);
                }
                for (iTel = 0; iTel < nbTel; iTel++) 
                    /* We can loop on telescopes here as long as the
                     * interf[] gets light from all telescopes, and the
                     * order of the operation is not important */
                {
                    mkPtr[iFrame][channel][iPix] -= 
                        phot[iFrame][channel][iTel] * vk[iTel][iPix];

                    sigma2MkPtr[iFrame][channel][iPix] +=
                        sigma2Phot[iFrame][channel][iTel] *
                        amdlibPow2(vk[iTel][iPix]);
                }
            }
            amdlibFree2DArrayFloatWrapping(intfPtr);
            amdlibFree2DArrayFloatWrapping(intfSkyPtr);
        }

        /* Store corresponding wavelength in amdlibWAVELENGTH structure*/
        wave->bandwidth[channel] = waveData->bandwidth[channelNo[channel]];
        wave->wlen[channel]      = p2vm->wlen[iP2vmChannel];
    }
    /* Set number of channels */
    wave->nbWlen = nbChannels;

    amdlibCrossP2vmAndData_FREEALL();

    return amdlibSUCCESS;
}
#undef amdlibCrossP2vmAndData_FREEALL

/** Useful macro to free amdlibInvertP2v() structures */ 
#define amdlibInvertP2v_FREEALL()                     \
    free(mat3); free(mat2); free(mat1); free(tp2v);   \
    amdlibFree2DArrayDouble(invcovmat);               \
    amdlibFree2DArrayUnsignedCharWrapping(badp2vPtr); \
    amdlibFree3DArrayDoubleWrapping(p2vPtr);          \
    amdlibFree3DArrayDoubleWrapping(sigma2_mkPtr);    \
    amdlibFree3DArrayDoubleWrapping(v2pPtr);
/**
 * Invert p2vm. 
 * 
 * @param p2v p2vm to invert.
 * @param badp2v part of bad pixels map related to the p2vm.
 * @param sigma2_mk the variance of the continuum-corrected interferograms 
 * (needed to estimate the covariance matrix).
 * @param binning number of frames in the current binning.
 * @param nbChannels number of valid spectral channels treated.
 * @param nbPix number of pixels (width of p2vm).
 * @param nbBases number of baselines.
 * @param v2p output p2vm computed: invert of p2v.
 * @param errMsg error description message returned if function fails.
 *
 * @return 
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.
 */
amdlibCOMPL_STAT amdlibInvertP2v(/* Input */
                                 double          *p2v,
                                 unsigned char   *badp2v,
                                 double          *sigma2_mk,
                                 int             binning,
                                 int             nbChannels,
                                 int             nbPix,
                                 int             nbBases,
                                 /* Output */
                                 double          *v2p,
                                 amdlibERROR_MSG errMsg)
{
    int iPix; /*usually loop on nbPix*/
    int iFrame; /* usually loop on number of frames */
    int lVis;

    double *tp2v = NULL; /* p2v's transposed matrix */
    double *mat1 = NULL;
    double *mat2 = NULL;
    double *mat3 = NULL;
    double **invcovmat = NULL;

    double ***p2vPtr = NULL;       /* Wrap on input p2v */
    double ***sigma2_mkPtr = NULL; /* Wrap on input sigma2_mk */
    double ***v2pPtr = NULL;       /* Wrap on output v2p */
    unsigned char **badp2vPtr = NULL;


    amdlibLogTrace("amdlibInvertP2v()");

    tp2v = calloc(2*nbPix*nbBases, sizeof(double));
    if (tp2v == NULL)
    {
        amdlibSetErrMsg("Allocation of tp2v failed");
        amdlibInvertP2v_FREEALL();
        return amdlibFAILURE;
    }
    mat1 = calloc(2*nbBases*nbPix, sizeof(double));
    if (mat1 == NULL)
    {
        amdlibSetErrMsg("Allocation of mat1 failed");
        amdlibInvertP2v_FREEALL();
        return amdlibFAILURE;
    }
    mat2 = calloc(2*nbBases*2*nbBases, sizeof(double));
    if (mat2 == NULL)
    {
        amdlibSetErrMsg("Allocation of mat2 failed");
        amdlibInvertP2v_FREEALL();
        return amdlibFAILURE;
    }
    mat3 = calloc(2*nbPix*nbBases, sizeof(double));
    if (mat3 == NULL)
    {
        amdlibSetErrMsg("Allocation of mat3 failed");
        amdlibInvertP2v_FREEALL();
        return amdlibFAILURE;
    }
    invcovmat = amdlibAlloc2DArrayDouble(nbPix, nbPix, errMsg);
    if (invcovmat == NULL)
    {
        amdlibInvertP2v_FREEALL();
        return amdlibFAILURE;
    }

    p2vPtr = amdlibWrap3DArrayDouble(p2v, nbPix, 2*nbBases, nbChannels, errMsg);
    if (p2vPtr == NULL)
    {
        amdlibInvertP2v_FREEALL();
        return amdlibFAILURE;
    }
    sigma2_mkPtr = amdlibWrap3DArrayDouble(sigma2_mk, nbPix, nbChannels, 
                                           binning, errMsg);
    if (sigma2_mkPtr == NULL)
    {
        amdlibInvertP2v_FREEALL();
        return amdlibFAILURE;
    }
    v2pPtr = amdlibWrap3DArrayDouble(v2p, nbPix, 2*nbBases, nbChannels, errMsg);
    if (v2pPtr == NULL)
    {
        amdlibInvertP2v_FREEALL();
        return amdlibFAILURE;
    }
    badp2vPtr = amdlibWrap2DArrayUnsignedChar(badp2v, nbPix, nbChannels, errMsg);
    if (badp2vPtr == NULL)
    {
        amdlibInvertP2v_FREEALL();
        return amdlibFAILURE;
    }


    for (lVis=0; lVis < nbChannels; lVis++)
    {
        memset(invcovmat[0], '\0', nbPix*nbPix*sizeof(double));

        /* We get rid of the bad Pixels in the p2v matrix in the
         * following way: During the computation of the matrix, the ck
         * and dk of a bad pixel k are set to a value of 1 and zero
         * respectively. The weight of the bad pixel will be zero in
         * the covariance matrix and this should suffice to avoid
         * using it in the p2v. However, when a complete row of a p2v
         * matrix is zero, the following will crash. Fortunately we
         * prevented that above by selecting only "good" rows of p2v .
         */
        for (iPix = 0; iPix < nbPix; iPix++)
        {

            if (badp2vPtr[lVis][iPix] == amdlibGOOD_PIXEL_FLAG)
            {
                double tmp = 0.0;
                for (iFrame = 0; iFrame < binning; iFrame++)
                {
                    tmp += sigma2_mkPtr[iFrame][lVis][iPix];
                }

                tmp = tmp / binning;
                invcovmat[iPix][iPix] = 1./tmp;
            }
            else
            {
                invcovmat[iPix][iPix] = 0.0;
            }
        }
        /* computing the v2p (inverse of p2v) of this lambda, averaged
         * over all frames */
        /* following notation of report amb-igr-xx:
         * compute mat1 = cov^-1 x p2v
         * compute mat2 = tp2v x cov^-1 x p2v
         * then mat2 = (mat2)^-1
         * then mat3 = tp2v cov^-1
         * the v2p being
         * (tp2v x cov^-1 x p2v)^-1 x tp2v x cov^-1
         * is also mat2 x mat3
         */
        if (amdlibProductMatrix(invcovmat[0], p2vPtr[lVis][0], mat1, nbPix, 
                                nbPix, 2*nbBases) != amdlibSUCCESS)
        {
            amdlibSetErrMsg("Product matrix error.");
            amdlibInvertP2v_FREEALL();
            return amdlibFAILURE;
        }

        amdlibTransposeMatrix(p2vPtr[lVis][0], tp2v, nbPix, 2*nbBases);

        if (amdlibProductMatrix(tp2v, mat1, mat2, 2*nbBases, nbPix, 
                                2*nbBases) != amdlibSUCCESS)
        {
            amdlibSetErrMsg("Product matrix error.");
            amdlibInvertP2v_FREEALL();
            return amdlibFAILURE;
        }

        amdlibInvertMatrix(mat2, 2*nbBases);

        if (amdlibProductMatrix(mat2, tp2v, mat3, 2*nbBases, 2*nbBases, 
                                nbPix)!= amdlibSUCCESS)
        {
            amdlibSetErrMsg("Product matrix error.");
            amdlibInvertP2v_FREEALL();
            return amdlibFAILURE;
        }
        if (amdlibProductMatrix(mat3, invcovmat[0], v2pPtr[lVis][0], 2*nbBases,
                                nbPix, nbPix) != amdlibSUCCESS)
        {
            amdlibSetErrMsg("Product matrix error.");
            amdlibInvertP2v_FREEALL();
            return amdlibFAILURE;
        }

    }
    amdlibInvertP2v_FREEALL();
    return amdlibSUCCESS;
}
#undef amdlibInvertP2v_FREEALL

/** Useful macro to free amdlibComputeCorrFlux() structures */ 
#define amdlibComputeCorrFlux_FREEALL()             \
    amdlibFree3DArrayDoubleWrapping(mkPtr);         \
    amdlibFree3DArrayDoubleWrapping(sigma2_mkPtr);  \
    amdlibFree3DArrayDoubleWrapping(v2pPtr);        \
    amdlibFree2DArrayWrapping((void **)tablePtr);
/**
 * Compute correlated flux. 
 * 
 * @param v2p input p2vm used.
 * @param mk  continumm-corrected interferograms.
 * @param sigma2_mk the associated variances.
 * @param binning number of frames in the current binning.
 * @param nbLVis number of spectral channels treated.
 * @param nbPix number of pixels (width of p2vm).
 * @param nbBases number of baselines.
 * @param vis (instant vis) structure were results are stored.
 * @param errMsg error description message returned if function fails.
 *
 * @return 
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.
 */
amdlibCOMPL_STAT amdlibComputeCorrFlux(/* Input */
                                       double                 *v2p,
                                       double                 *mk,
                                       double                 *sigma2_mk,
                                       int                    binning,
                                       int                    nbLVis,
                                       int                    nbPix,
                                       int                    nbBases,
                                       /* Output */
                                       amdlibVIS              *vis,
                                       amdlibERROR_MSG        errMsg)
{

    /* loop indexes and associated limits */
    int iPix; /*usually loop on nbPix*/
    int iFrame; /* usually loop on number of frames */
    int iBase; /*usually loop on nbBases*/
    int lVis = 0;

    amdlibCOMPLEX CorrFlux;
    amdlibCOMPLEX sigma2_CorrFlux;

    double        cov_RI;

    double ***mkPtr = NULL; /* Wrap on input mk */
    double ***sigma2_mkPtr = NULL; /* Wrap on input sigma2_mk */
    double ***v2pPtr = NULL;       /* Wrap on input v2p */
    amdlibVIS_TABLE_ENTRY **tablePtr = NULL; /* Wrap on output vis->table */

    amdlibLogTrace("amdlibComputeCorrFlux()");

    mkPtr = amdlibWrap3DArrayDouble(mk, nbPix, nbLVis, binning, errMsg);
    if (mkPtr == NULL)
    {
        amdlibComputeCorrFlux_FREEALL();
        return amdlibFAILURE;
    }
    sigma2_mkPtr = amdlibWrap3DArrayDouble(sigma2_mk, nbPix, nbLVis, 
                                           binning, errMsg);
    if (sigma2_mkPtr == NULL)
    {
        amdlibComputeCorrFlux_FREEALL();
        return amdlibFAILURE;
    }
    v2pPtr = amdlibWrap3DArrayDouble(v2p, nbPix, 2*nbBases, nbLVis, errMsg);
    if (v2pPtr == NULL)
    {
        amdlibComputeCorrFlux_FREEALL();
        return amdlibFAILURE;
    }
    tablePtr = 
        (amdlibVIS_TABLE_ENTRY **)amdlibWrap2DArray(vis->table,
                                                    nbBases, vis->nbFrames,
                                                    sizeof(amdlibVIS_TABLE_ENTRY),
                                                    errMsg);
    if (tablePtr == NULL)
    {
        amdlibComputeCorrFlux_FREEALL();
        return amdlibFAILURE;
    }

    for (lVis = 0; lVis < nbLVis; lVis++)
    {
        for (iFrame = 0; iFrame < binning; iFrame++)
        {
            for (iBase = 0; iBase < nbBases; iBase++)
            { /* begin loop for iBase */

                /* Initialization of the useful data */
                CorrFlux.re = 0.0;
                CorrFlux.im = 0.0;
                sigma2_CorrFlux.re = 0.0;
                sigma2_CorrFlux.im = 0.0;
                cov_RI = 0.0;

                /* The matrix-vector product */
                for (iPix = 0; iPix < nbPix; iPix++)
                {
                    /* calculation of CorrFlux (eq 33) and
                     * save those CorrFlux into table 'vis'
                     */
                    CorrFlux.re += v2pPtr[lVis][iBase][iPix] * 
                        mkPtr[iFrame][lVis][iPix];

                    CorrFlux.im += v2pPtr[lVis][iBase+nbBases][iPix] * 
                        mkPtr[iFrame][lVis][iPix];

                    /* calculation of sigma^2 of CorrFlux
                     * and the COV(R,I) see eqs 44, 45 and 46
                     * used in bias see formula AMB-IGR-019 number 36
                     */

                    /*eq 44*/
                    sigma2_CorrFlux.re += v2pPtr[lVis][iBase][iPix] * 
                        v2pPtr[lVis][iBase][iPix] * 
                        sigma2_mkPtr[iFrame][lVis][iPix];

                    /*eq 45*/
                    sigma2_CorrFlux.im += v2pPtr[lVis][iBase+nbBases][iPix] *
                        v2pPtr[lVis][iBase+nbBases][iPix] * 
                        sigma2_mkPtr[iFrame][lVis][iPix];

                    cov_RI += v2pPtr[lVis][iBase][iPix] * 
                        v2pPtr[lVis][iBase+nbBases][iPix] *
                        sigma2_mkPtr[iFrame][lVis][iPix];

                }

                /* Store For time being Raw Correlated Flux in table Vis */

                tablePtr[iFrame][iBase].vis[lVis].re = CorrFlux.re;

                tablePtr[iFrame][iBase].vis[lVis].im = CorrFlux.im;

                tablePtr[iFrame][iBase].sigma2Vis[lVis].re = sigma2_CorrFlux.re;

                tablePtr[iFrame][iBase].sigma2Vis[lVis].im = sigma2_CorrFlux.im;

                tablePtr[iFrame][iBase].visCovRI[lVis] = cov_RI;
            }
        }
    }
    amdlibComputeCorrFlux_FREEALL();
    return amdlibSUCCESS;
}
#undef amdlibComputeCorrFlux_FREEALL

/** Useful macro to free amdlibSumPhotometry() structures */ 
#define amdlibSumPhotometry_FREEALL()                        \
    amdlibFree2DArrayWrapping((void **)instantPhotTablePtr); \
    amdlibFree2DArrayWrapping((void **)summedPhotTablePtr); \
    amdlibFree3DArrayDouble(tmpphot);                 \
    amdlibFree3DArrayDouble(s2phot);                  
/**
 * Sum photometry. 
 * 
 * @param instantPhot photometry to sum.
 * @param iBin index of current binning.
 * @param summedPhot photometry structure were results are stored.
 *
 * @return 
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.
 */
amdlibCOMPL_STAT amdlibSumPhotometry(amdlibPHOTOMETRY *instantPhot,
                                     int              iBin,
                                     amdlibPHOTOMETRY *summedPhot)
{
    int iFrame; /* usually loop on number of frames */
    int nbFrames = instantPhot->nbFrames;
    int iBase; /*usually loop on nbBases*/
    int nbBases = instantPhot->nbBases;
    int lPhot = 0;
    int nbLPhot = instantPhot->nbWlen;

    double ***tmpphot=NULL,***s2phot=NULL;
    double P1MoreP2, P1DiviP2, P1, P2,sigma2P1MoreP2,sigma2P1DiviP2,  sigma2_P1, sigma2_P2;

    amdlibPHOTOMETRY_TABLE_ENTRY **instantPhotTablePtr = NULL;
    amdlibPHOTOMETRY_TABLE_ENTRY **summedPhotTablePtr = NULL;
    static amdlibERROR_MSG errMsg;

    instantPhotTablePtr = 
        (amdlibPHOTOMETRY_TABLE_ENTRY **)amdlibWrap2DArray(instantPhot->table,
                                        nbBases, nbFrames,
                                        sizeof(amdlibPHOTOMETRY_TABLE_ENTRY),
                                        errMsg);
    if (instantPhotTablePtr == NULL)
    {
        amdlibSumPhotometry_FREEALL();
        return amdlibFAILURE;
    }
    summedPhotTablePtr = 
        (amdlibPHOTOMETRY_TABLE_ENTRY **)amdlibWrap2DArray(summedPhot->table,
                                        nbBases, summedPhot->nbFrames,
                                        sizeof(amdlibPHOTOMETRY_TABLE_ENTRY),
                                        errMsg);
    if (summedPhotTablePtr == NULL)
    {
        amdlibSumPhotometry_FREEALL();
        return amdlibFAILURE;
    }

    /* Allocate memory for retrieval of individual (per-beam) photometries
     * prior to summation of per-beam photometries and errors before
     * reconversion to Sum and Ratio of Fluxes and Errors*/
    tmpphot = amdlibAlloc3DArrayDouble(2, nbLPhot, nbBases, errMsg);
    s2phot = amdlibAlloc3DArrayDouble(2, nbLPhot, nbBases, errMsg);

    amdlibLogTrace("amdlibSumPhotometry()");
    
    for (lPhot = 0; lPhot < nbLPhot; lPhot++)
    {
        for (iBase = 0; iBase < nbBases; iBase++)
        {
            for (iFrame = 0; iFrame < nbFrames; iFrame++)
            {
                P1MoreP2 = 
                    instantPhotTablePtr[iFrame][iBase].fluxSumPiPj[lPhot];
                sigma2P1MoreP2 = 
                    instantPhotTablePtr[iFrame][iBase].sigma2FluxSumPiPj[lPhot];
                P1DiviP2 = 
                    instantPhotTablePtr[iFrame][iBase].fluxRatPiPj[lPhot];
                sigma2P1DiviP2 = 
                    instantPhotTablePtr[iFrame][iBase].sigma2FluxRatPiPj[lPhot];

                P2 = P1MoreP2 / (1.0 + P1DiviP2) ;
                sigma2_P2 = sigma2P1MoreP2 / (1.0 + sigma2P1DiviP2) ;

                P1 = P1MoreP2*P1DiviP2/(1.0+P1DiviP2);
                sigma2_P1 = sigma2P1MoreP2*sigma2P1DiviP2/(1.0+sigma2P1DiviP2);

                tmpphot[iBase][lPhot][0] += P1;
                tmpphot[iBase][lPhot][1] += P2;

                s2phot[iBase][lPhot][0] += sigma2_P1;
                s2phot[iBase][lPhot][1] += sigma2_P2;

                summedPhotTablePtr[iBin][iBase].PiMultPj[lPhot] +=
                instantPhotTablePtr[iFrame][iBase].PiMultPj[lPhot];
            }
        }
    }
    for (lPhot = 0; lPhot < nbLPhot; lPhot++)
    {
        for (iBase = 0; iBase < nbBases; iBase++)
        {
            summedPhotTablePtr[iBin][iBase].fluxSumPiPj[lPhot]=
            tmpphot[iBase][lPhot][0]+tmpphot[iBase][lPhot][1];
            summedPhotTablePtr[iBin][iBase].sigma2FluxSumPiPj[lPhot]=
            s2phot[iBase][lPhot][0]+s2phot[iBase][lPhot][1];
            summedPhotTablePtr[iBin][iBase].fluxRatPiPj[lPhot]=
            tmpphot[iBase][lPhot][0]/tmpphot[iBase][lPhot][1];
            summedPhotTablePtr[iBin][iBase].sigma2FluxRatPiPj[lPhot]=
            s2phot[iBase][lPhot][0]/s2phot[iBase][lPhot][1];
        }
    }
    
    amdlibSumPhotometry_FREEALL();
    return amdlibSUCCESS;
}
#undef amdlibSumPhotometry_FREEALL

/** Useful macro to free amdlibMeanCpxVis() structures */ 
#define amdlibMeanCpxVis_FREEALL()                         \
    free(sigma2_normedVis_im); free(sigma2_normedVis_re);  \
    free(normedVis_im); free(normedVis_re);                \
    amdlibFree2DArrayWrapping((void **)corrFluxTablePtr);  \
    amdlibFree2DArrayWrapping((void **)photTablePtr);      \
    amdlibFree2DArrayWrapping((void **)normedVisTablePtr);
/**
 * Compute averaged, normalized visibilites from a sequence (bin) of
 * instanteous correlated fluxes. 
 * 
 * @param instantCorrFlux the bin of  instanteous correlated fluxes.
 * @param photometry the associated bin of instantenous fluxes.
 * @param iBin index of the results to be written in the output (binned) 
 * normedVis structure.
 * @param normedVis structure were results are stored.
 *
 * @return 
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.
 */
amdlibCOMPL_STAT amdlibMeanCpxVis(amdlibVIS         *instantCorrFlux,
                                  amdlibPHOTOMETRY  *photometry,
                                  int               iBin,
                                  amdlibVIS         *normedVis)
{
    int iFrame; /* usually loop on number of frames */
    int nbFrames = instantCorrFlux->nbFrames;
    int iBase; /*usually loop on nbBases*/
    int nbBases = instantCorrFlux->nbBases;

    int lVis = 0;
    int nbLVis = instantCorrFlux->nbWlen;

    amdlibCOMPLEX CorrFlux, sigma2_CorrFlux;

    double P1P2;
    double *normedVis_re = NULL;
    double *normedVis_im = NULL;
    double *sigma2_normedVis_re = NULL;
    double *sigma2_normedVis_im = NULL;

    amdlibVIS_TABLE_ENTRY        **corrFluxTablePtr = NULL;
    amdlibPHOTOMETRY_TABLE_ENTRY **photTablePtr = NULL;
    amdlibVIS_TABLE_ENTRY        **normedVisTablePtr = NULL;
    static amdlibERROR_MSG errMsg;


    amdlibLogTrace("amdlibMeanCpxVis()");

    /* Allocate memory for intermediate structures */
    normedVis_re = calloc(nbFrames,sizeof(double));
    if (normedVis_re == NULL)
    {
        amdlibMeanCpxVis_FREEALL();
        return amdlibFAILURE;
    }
    normedVis_im = calloc(nbFrames,sizeof(double));
    if (normedVis_im == NULL)
    {
        amdlibMeanCpxVis_FREEALL();
        return amdlibFAILURE;
    }
    sigma2_normedVis_re = calloc(nbFrames,sizeof(double));
    if (sigma2_normedVis_re == NULL)
    {
        amdlibMeanCpxVis_FREEALL();
        return amdlibFAILURE;
    }
    sigma2_normedVis_im = calloc(nbFrames,sizeof(double));
    if (sigma2_normedVis_im == NULL)
    {
        amdlibMeanCpxVis_FREEALL();
        return amdlibFAILURE;
    }

    /* And wrap structures specified as parametres */
    corrFluxTablePtr = 
        (amdlibVIS_TABLE_ENTRY **)amdlibWrap2DArray(instantCorrFlux->table,
                                                nbBases, nbFrames,
                                                sizeof(amdlibVIS_TABLE_ENTRY),
                                                errMsg);
    if (corrFluxTablePtr == NULL)
    {
        amdlibMeanCpxVis_FREEALL();
        return amdlibFAILURE;
    }
    photTablePtr = 
        (amdlibPHOTOMETRY_TABLE_ENTRY **)amdlibWrap2DArray(photometry->table,
                                        nbBases, nbFrames,
                                        sizeof(amdlibPHOTOMETRY_TABLE_ENTRY),
                                        errMsg);
    if (photTablePtr == NULL)
    {
        amdlibMeanCpxVis_FREEALL();
        return amdlibFAILURE;
    }
    normedVisTablePtr = 
        (amdlibVIS_TABLE_ENTRY **)amdlibWrap2DArray(normedVis->table,
                                                nbBases, normedVis->nbFrames,
                                                sizeof(amdlibVIS_TABLE_ENTRY),
                                                errMsg);
    if (normedVisTablePtr == NULL)
    {
        amdlibMeanCpxVis_FREEALL();
        return amdlibFAILURE;
    }

    /* Compute averaged, normalized visibilites */
    for (lVis = 0; lVis < nbLVis; lVis++)
    { 
        for (iBase = 0; iBase < nbBases; iBase++)
        {
            for (iFrame = 0; iFrame < nbFrames; iFrame++)
            {
                P1P2 = photTablePtr[iFrame][iBase].PiMultPj[lVis];

                CorrFlux.re = corrFluxTablePtr[iFrame][iBase].vis[lVis].re;
                CorrFlux.im = corrFluxTablePtr[iFrame][iBase].vis[lVis].im;

                sigma2_CorrFlux.re = 
                    corrFluxTablePtr[iFrame][iBase].sigma2Vis[lVis].re;
                sigma2_CorrFlux.im =
                    corrFluxTablePtr[iFrame][iBase].sigma2Vis[lVis].im;

                normedVis_re[iFrame] = CorrFlux.re / 
                    (2 * amdlibSignedSqrt(P1P2));
                normedVis_im[iFrame] = CorrFlux.im / 
                    (2 * amdlibSignedSqrt(P1P2));
                sigma2_normedVis_re[iFrame] = sigma2_CorrFlux.re / (4 * P1P2);
                sigma2_normedVis_im[iFrame] = sigma2_CorrFlux.im / (4 * P1P2);

            }
            /* Store results */
            normedVisTablePtr[iBin][iBase].vis[lVis].re =
                amdlibAvgTable(nbFrames, normedVis_re, sigma2_normedVis_re);

            normedVisTablePtr[iBin][iBase].vis[lVis].im =
                amdlibAvgTable(nbFrames, normedVis_im, sigma2_normedVis_im);

            normedVisTablePtr[iBin][iBase].sigma2Vis[lVis].re =
                amdlibAvgTable(nbFrames, sigma2_normedVis_re, 
                               sigma2_normedVis_re);

            normedVisTablePtr[iBin][iBase].sigma2Vis[lVis].im =
                amdlibAvgTable(nbFrames, sigma2_normedVis_im, 
                               sigma2_normedVis_im);
        }
    }

    amdlibMeanCpxVis_FREEALL();
    return amdlibSUCCESS;
}
#undef amdlibMeanCpxVis_FREEALL

/** Useful macro to free amdlibComputeSqVis() structures */ 
#define amdlibComputeSqVis_FREEALL() free(sigma_V2);                     \
    amdlibFree2DArrayDouble(I); amdlibFree2DArrayDouble(sigma2_R);       \
    amdlibFree2DArrayDouble(sigma2_I); amdlibFree2DArrayDouble(cov_RI);  \
    amdlibFree2DArrayDouble(instantP1P2); free(averageP1P2);             \
    free(sigma2_P1P2); amdlibFree2DArrayDouble(unbiased_CV);             \
    free(average_CV); free(sigma2_CV); free(averageV2); free(sigma2_V2); \
    amdlibFree2DArrayDouble(R); free(theoretical_averageP1P2);           \
    amdlibFree2DArrayWrapping((void **)corrFluxTablePtr);                \
    amdlibFree2DArrayWrapping((void **)photTablePtr);                    \
    amdlibFree2DArrayWrapping((void **)vis2TablePtr);

/**
 * Compute averaged, normalized, debiassed, squared visibilities from a 
 * sequence (bin) of instanteous correlated fluxes. 
 * 
 * @param instantCorrFlux the bin of  instanteous correlated fluxes.
 * @param photometry the associated bin of instantenous fluxes.
 * @param iBin index of the results to be written in the output 
 * (binned) vis2 structure.
 * @param band the current spectral band treated.
 * @param selection structure containing all data relative to frame selection.
 * @param errorType indicates wether the noise figures are estimated 
 * statistically from the sequence of photometries during the bin 
 * (amdlibSTATISTICAL_ERROR) or theoretically by the poisson error on N photons
 * (amdlibTHEORETICAL_ERROR). The latter is forced obviously when the binsize
 * is 1.
 * @param vis2 structure were results are stored.
 * @param errMsg error description message returned if function fails.
 *
 * @return 
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.
 */
amdlibCOMPL_STAT amdlibComputeSqVis(/* Input */
                                    amdlibVIS         *instantCorrFlux,
                                    amdlibPHOTOMETRY  *photometry,
                                    int               iBin,
                                    amdlibBAND        band,
                                    amdlibSELECTION   *selection,
                                    amdlibERROR_TYPE  errorType,
                                    /* Output */
                                    amdlibVIS2        *vis2,
                                    amdlibERROR_MSG   errMsg)
{

    int iFrame, iSelFram; /* usually loop on number of frames */
    int nbFrames = instantCorrFlux->nbFrames;
    int iBase; /*usually loop on nbBases*/
    int nbBases = instantCorrFlux->nbBases;
    int lVis = 0;
    int nbLVis = instantCorrFlux->nbWlen;

    double P1MoreP2, P1DiviP2, sigma2P1MoreP2, sigma2P1DiviP2;
    double P1;
    double P2;
    double sigma2_P1;
    double sigma2_P2;
    double biased_CV;
    double bias;

    double **R = NULL;
    double **I = NULL;
    double **sigma2_R = NULL;
    double **sigma2_I = NULL;
    double **cov_RI = NULL;
    double **instantP1P2 = NULL;
    double *theoretical_averageP1P2 = NULL;
    double *averageP1P2 = NULL;
    double *sigma2_P1P2 = NULL;
    double **unbiased_CV = NULL;
    double *average_CV = NULL;
    double *sigma2_CV = NULL;
    double *averageV2 = NULL;
    double *sigma2_V2 = NULL;
    double *sigma_V2 = NULL;

    amdlibVIS_TABLE_ENTRY **corrFluxTablePtr = NULL;
    amdlibPHOTOMETRY_TABLE_ENTRY **photTablePtr = NULL;
    amdlibVIS2_TABLE_ENTRY **vis2TablePtr = NULL;

    amdlibLogTrace("amdlibComputeSqVis()");

    /* Allocate memory for intermediate structures, of maximum size in frames */
    R = amdlibAlloc2DArrayDouble(nbBases, nbFrames, errMsg);
    if (R == NULL)
    {
        amdlibComputeSqVis_FREEALL();
        return amdlibFAILURE;
    }
    I = amdlibAlloc2DArrayDouble(nbBases, nbFrames, errMsg);
    if (I == NULL)
    {
        amdlibComputeSqVis_FREEALL();
        return amdlibFAILURE;
    }
    sigma2_R = amdlibAlloc2DArrayDouble(nbBases, nbFrames, errMsg);
    if (sigma2_R == NULL)
    {
        amdlibComputeSqVis_FREEALL();
        return amdlibFAILURE;
    }
    sigma2_I = amdlibAlloc2DArrayDouble(nbBases, nbFrames, errMsg);
    if (sigma2_I == NULL)
    {
        amdlibComputeSqVis_FREEALL();
        return amdlibFAILURE;
    }
    cov_RI = amdlibAlloc2DArrayDouble(nbBases, nbFrames, errMsg);
    if (cov_RI == NULL)
    {
        amdlibComputeSqVis_FREEALL();
        return amdlibFAILURE;
    }
    instantP1P2 = amdlibAlloc2DArrayDouble(nbFrames, nbBases, errMsg);          
    if (instantP1P2 == NULL)
    {
        amdlibComputeSqVis_FREEALL();
        return amdlibFAILURE;
    }
    averageP1P2 = calloc(nbBases, sizeof(double));
    if (averageP1P2 == NULL)
    {
        amdlibComputeSqVis_FREEALL();
        amdlibSetErrMsg("Allocation of averageP1P2 failed");
        return amdlibFAILURE;
    }
    theoretical_averageP1P2 = calloc(nbBases, sizeof(double));
    if (theoretical_averageP1P2 == NULL)
    {
        amdlibComputeSqVis_FREEALL();
        amdlibSetErrMsg("Allocation of theoretical_averageP1P2 failed");
        return amdlibFAILURE;
    }
    sigma2_P1P2 = calloc(nbBases,sizeof(double));
    if (sigma2_P1P2 == NULL)
    {
        amdlibComputeSqVis_FREEALL();
        amdlibSetErrMsg("Allocation of sigma2_P1P2 failed");
        return amdlibFAILURE;
    }
    unbiased_CV = amdlibAlloc2DArrayDouble(nbFrames, nbBases, errMsg);
    if (unbiased_CV == NULL)
    {
        amdlibComputeSqVis_FREEALL();
        return amdlibFAILURE;
    }
    average_CV = calloc(nbBases,sizeof(double));
    if (average_CV == NULL)
    {
        amdlibComputeSqVis_FREEALL();
        amdlibSetErrMsg("Allocation of average_CV failed");
        return amdlibFAILURE;
    }
    sigma2_CV = calloc(nbBases,sizeof(double));
    if (sigma2_CV == NULL)
    {
        amdlibComputeSqVis_FREEALL();
        amdlibSetErrMsg("Allocation of sigma2_CV failed");
        return amdlibFAILURE;
    }
    averageV2 = calloc(nbBases,sizeof(double));
    if (averageV2 == NULL)
    {
        amdlibComputeSqVis_FREEALL();
        amdlibSetErrMsg("Allocation of averageV2 failed");
        return amdlibFAILURE;
    }
    sigma2_V2 = calloc(nbBases,sizeof(double));
    if (sigma2_V2 == NULL)
    {
        amdlibComputeSqVis_FREEALL();
        amdlibSetErrMsg("Allocation of sigma2_V2 failed");
        return amdlibFAILURE;
    }
    sigma_V2 = calloc(nbBases,sizeof(double));
    if (sigma_V2 == NULL)
    {
        amdlibComputeSqVis_FREEALL();
        amdlibSetErrMsg("Allocation of sigma_V2 failed");
        return amdlibFAILURE;
    }

    /* Wrap structures given as parametres */
    corrFluxTablePtr = 
        (amdlibVIS_TABLE_ENTRY **)amdlibWrap2DArray(instantCorrFlux->table,
                                            nbBases, nbFrames,
                                            sizeof(amdlibVIS_TABLE_ENTRY),
                                            errMsg);
    if (corrFluxTablePtr == NULL)
    {
        amdlibComputeSqVis_FREEALL();
        return amdlibFAILURE;
    }
    photTablePtr = 
        (amdlibPHOTOMETRY_TABLE_ENTRY **)amdlibWrap2DArray(photometry->table,
                                nbBases, photometry->nbFrames,
                                sizeof(amdlibPHOTOMETRY_TABLE_ENTRY),
                                errMsg);
    if (photTablePtr == NULL)
    {
        amdlibComputeSqVis_FREEALL();
        return amdlibFAILURE;
    }
    vis2TablePtr = 
        (amdlibVIS2_TABLE_ENTRY **)amdlibWrap2DArray(vis2->table,
                                                nbBases, vis2->nbFrames,
                                                sizeof(amdlibVIS2_TABLE_ENTRY),
                                                errMsg);
    if (vis2TablePtr == NULL)
    {
        amdlibComputeSqVis_FREEALL();
        return amdlibFAILURE;
    }

    /* Do the work per wavelength, then per base (since amount of 
     * selected frames is baseline-dependent). */
    for (lVis = 0; lVis < nbLVis; lVis++)
    {
        /* SELECTED Frames: computation of the Squared Coherent Flux 
         * (at current wavelength) */
        for (iBase = 0; iBase < nbBases; iBase++)
        {
            for (iFrame = 0, iSelFram=0; iFrame < nbFrames; iFrame++)
            {
                if (selection->band[band].isSelectedPt[iBase][iFrame] == amdlibTRUE)
                {
                    R[iSelFram][iBase] =
                        corrFluxTablePtr[iFrame][iBase].vis[lVis].re;

                    I[iSelFram][iBase] =
                        corrFluxTablePtr[iFrame][iBase].vis[lVis].im;

                    sigma2_R[iSelFram][iBase] = 
                        corrFluxTablePtr[iFrame][iBase].sigma2Vis[lVis].re;

                    sigma2_I[iSelFram][iBase] = 
                        corrFluxTablePtr[iFrame][iBase].sigma2Vis[lVis].im;

                    cov_RI[iSelFram][iBase] = 
                        corrFluxTablePtr[iFrame][iBase].visCovRI[lVis];

                    /* computation of the unbiased, weighted squared
                     * visibilities
                     * Clm^2 unbiased_CV: records C^2(iFrame) for each frame for
                     * this wavelength we need this expression to calculate
                     * sigma_V2
                     */
                    biased_CV = amdlibPow2(R[iSelFram][iBase]) + 
                        amdlibPow2(I[iSelFram][iBase]);

                    /* bias: see formula AMB-IGR-019 number 36 */
                    /* for every baseline, iframe */   
                    bias = sigma2_R[iSelFram][iBase] + 
                        sigma2_I[iSelFram][iBase];

                    /*
                     * <R^2 + I^2 - bias>_fram =
                     * <R^2>_fram + <I^2>_fram - <bias>_fram averaged =
                     * <C^2 - bias>
                     */
                    unbiased_CV[iBase][iSelFram] = biased_CV - bias;

                    /* computation of instantP1P2 for every frame/baseline for 
                     * this wavelength. We need this expression to calculate 
                     * sigma_V2 below.  Remember also averageP1P2, the average 
                     * photometry as defined below */

                    instantP1P2[iBase][iSelFram] =
                        photTablePtr[iFrame][iBase].PiMultPj[lVis];

                    /* increment subset counter */
                    iSelFram++;
                }
            }
        }
        /* On selected frames only: average computation */
        for (iBase = 0; iBase < nbBases; iBase++)
        {
            /* Zero averageP1P2 & average_CV */ 
            averageP1P2[iBase] = 0.0;          
            average_CV[iBase] = 0.0;
            for (iSelFram = 0; iSelFram < selection->band[band].nbSelectedFrames[iBase]; 
                 iSelFram++)
            {
                averageP1P2[iBase] += instantP1P2[iBase][iSelFram];
                average_CV[iBase] += unbiased_CV[iBase][iSelFram];
            }
            /* compute average of P1P2 & CV */
            averageP1P2[iBase] /= selection->band[band].nbSelectedFrames[iBase];
            average_CV[iBase] /= selection->band[band].nbSelectedFrames[iBase];

            /* computation of the average of V2 for this wavelength */
            averageV2[iBase] = average_CV[iBase] / (4 * averageP1P2[iBase]);
        }

        /* For statistical errors we need to have only the selected values 
         * as contiguous tables */
        for (iBase = 0; iBase < nbBases; iBase++)
        {
            if ((errorType == amdlibSTATISTICAL_ERROR) &&
                (selection->band[band].nbSelectedFrames[iBase] > 1))
            {
                /* estimate sigma2_V2 as in Bram's formula 40. Since size 
                 * of arrays change with the number of frames selected per 
                 * baseline, use amdlibComputeCov instead of 
                 * amdlibComputeMatrixCov */ 
                sigma2_CV[iBase] = amdlibComputeCov(unbiased_CV[iBase], 
                                            unbiased_CV[iBase], 
                                            selection->band[band].nbSelectedFrames[iBase]);

                sigma2_P1P2[iBase] = amdlibComputeCov(instantP1P2[iBase], 
                                            instantP1P2[iBase], 
                                            selection->band[band].nbSelectedFrames[iBase]);  
                
                /* computation of sigma_V2 */
                sigma2_V2[iBase] = amdlibPow2(averageV2[iBase]) / 
                    selection->band[band].nbSelectedFrames[iBase] * 
                    ((sigma2_CV[iBase] / amdlibPow2(average_CV[iBase])) +
                     (sigma2_P1P2[iBase] / amdlibPow2(averageP1P2[iBase])));

                if(sigma2_V2[iBase] >= 0)
                {
                    sigma_V2[iBase] = amdlibSignedSqrt(sigma2_V2[iBase]);
                }
                else
                {
                    sigma_V2[iBase] = amdlibBLANKING_VALUE;
                }
            }
            else  /* THEORETICAL ERROR . We have to recompute averageP1P2 
                   * also. reason: averageP1P2 above is an average of 
                   * P1*P2*sumvk. But sigma2_P1P2 retrieved below from P1+P2
                   * and P1/P2 does not know about sumvk. P1 is in fact 
                   * P1*sumvk1 and P2 is P2*sumvk2. So we recompute 
                   * averageP1P2 as the average of P1*P2, that goes with
                   * the theoretical sigma2_P1P2. */
            {
                theoretical_averageP1P2[iBase] = 0.0;
                sigma2_P1P2[iBase] = 0.0;
                sigma2_CV[iBase] = 0.0;
                for (iSelFram = 0; 
                     iSelFram < selection->band[band].nbSelectedFrames[iBase]; 
                     iSelFram++)
                { /* begin loop for iSelFram, selection has been done */
                    sigma2_CV[iBase] += 
                        4 * ((amdlibPow2(R[iSelFram][iBase])*
                              sigma2_R[iSelFram][iBase]) +
                             (amdlibPow2(I[iSelFram][iBase])*
                              sigma2_I[iSelFram][iBase]) +
                             2 * R[iSelFram][iBase] * I[iSelFram][iBase] * 
                             cov_RI[iSelFram][iBase]);
                }

                for (iFrame = 0; iFrame < nbFrames; iFrame++)
                { /* begin loop on iFrame, selection must be done here */
                    if (selection->band[band].isSelectedPt[iBase][iFrame] == amdlibTRUE)
                    {
                        P1MoreP2 = 
                            photTablePtr[iFrame][iBase].fluxSumPiPj[lVis];
                        sigma2P1MoreP2 = 
                            photTablePtr[iFrame][iBase].sigma2FluxSumPiPj[lVis];

                        P1DiviP2 = 
                            photTablePtr[iFrame][iBase].fluxRatPiPj[lVis];
                        sigma2P1DiviP2 = 
                            photTablePtr[iFrame][iBase].sigma2FluxRatPiPj[lVis];

                        P2 = P1MoreP2 / (1.0 + P1DiviP2);
                        
                        sigma2_P2 = sigma2P1MoreP2 / (1.0 + sigma2P1DiviP2);

                        P1 = P1MoreP2*P1DiviP2/(1.0+P1DiviP2);
                        
                        sigma2_P1 = sigma2P1MoreP2 * sigma2P1DiviP2 / 
                            (1.0 + sigma2P1DiviP2);
                        theoretical_averageP1P2[iBase] += P1*P2;

                        sigma2_P1P2[iBase] += amdlibPow2(P2) * sigma2_P1 + 
                            amdlibPow2(P1) * sigma2_P2;
                    }
                }
                /*all and all, the mean is on the selected frames only*/
                sigma2_P1P2[iBase] /= selection->band[band].nbSelectedFrames[iBase];
                sigma2_CV[iBase] /= selection->band[band].nbSelectedFrames[iBase];
                theoretical_averageP1P2[iBase] /= 
                    selection->band[band].nbSelectedFrames[iBase];

                sigma2_V2[iBase] = amdlibPow2(averageV2[iBase]) / 
                    selection->band[band].nbSelectedFrames[iBase] *
                    ((sigma2_CV[iBase] / amdlibPow2(average_CV[iBase])) +
                     (sigma2_P1P2[iBase] / 
                      amdlibPow2(theoretical_averageP1P2[iBase])));

                if(sigma2_V2[iBase] >= 0)
                {
                    sigma_V2[iBase] = amdlibSignedSqrt(sigma2_V2[iBase]);
                }
                else
                {
                    sigma_V2[iBase] = amdlibBLANKING_VALUE;
                }
            }            
        }

        for (iBase = 0; iBase < nbBases; iBase++)
        {
            if (selection->band[band].nbSelectedFrames[iBase] != 0)
            {
                vis2TablePtr[iBin][iBase].vis2[lVis] = averageV2[iBase];
                vis2TablePtr[iBin][iBase].vis2Error[lVis] = sigma_V2[iBase];
            }
        }
    }
    amdlibComputeSqVis_FREEALL();
    return amdlibSUCCESS;
}
#undef amdlibComputeSqVis_FREEALL

/** Useful macro to free amdlibAverageVis() structures */ 
#define amdlibAverageVisData_FREEALL()                         \
    amdlibFree2DArrayWrapping((void **)cpxVisTablePtr);        \
    amdlibFree2DArrayWrapping((void **)visTablePtr);   

/**
 * Update complex pseudo-visibilities in output table by averaging values and 
 * computing rms
 *
 * @param instantCpxVis the bin of  instanteous correlated fluxes.
 * @param band the current band
 * @param iBin index of the results to be written in the output 
 * (binned) vis structure.
 * @param selection the selection structure giving list of selected frames.
 * @param vis structure were results are stored.
 * @param errMsg error description message returned if function fails.
 *
 * @return 
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.
 */
amdlibCOMPL_STAT amdlibAverageVis(amdlibVIS              *instantCpxVis,
                                  int                    iBin,
                                  amdlibBAND             band,
                                  amdlibSELECTION        *selection,
                                  /* Output */
                                  amdlibVIS              *vis,
                                  amdlibERROR_MSG        errMsg)
{

    int iFrame; /* usually loop on number of frames */
    int nbFrames = instantCpxVis->nbFrames;
    int iBase; /*usually loop on nbBases*/
    int nbBases = instantCpxVis->nbBases;
    int lVis = 0;
    int nbLVis = instantCpxVis->nbWlen;

    amdlibVIS_TABLE_ENTRY **cpxVisTablePtr = NULL; 
    amdlibVIS_TABLE_ENTRY **visTablePtr = NULL; 
    amdlibCOMPLEX rmsVis, cpxVis;

    
    amdlibLogTrace("amdlibAverageVis()");

    /* Wrap structures given as parametres */
    cpxVisTablePtr = 
        (amdlibVIS_TABLE_ENTRY **)amdlibWrap2DArray(instantCpxVis->table,
                                                nbBases, nbFrames,
                                                sizeof(amdlibVIS_TABLE_ENTRY),
                                                errMsg);
    if (cpxVisTablePtr == NULL)
    {
        amdlibAverageVisData_FREEALL();
        return amdlibFAILURE;
    }

    visTablePtr = 
        (amdlibVIS_TABLE_ENTRY **)amdlibWrap2DArray(vis->table,
                                                nbBases, vis->nbFrames,
                                                sizeof(amdlibVIS_TABLE_ENTRY),
                                                errMsg);
    if (visTablePtr == NULL)
    {
        amdlibAverageVisData_FREEALL();
        return amdlibFAILURE;
    }
    /* Finally, update complex pseudo-visibilities in output table by averaging
     * values and computing rms. Reuse cpxVis and rmsVis for intermediate 
     * storage */
    for (lVis = 0; lVis < nbLVis; lVis++)
    {
        for (iBase = 0; iBase < nbBases; iBase++)
        {
            if (selection->band[band].nbSelectedFrames[iBase] != 0)
            {
                cpxVis.re = 0;
                cpxVis.im = 0;
                for (iFrame = 0; iFrame < nbFrames; iFrame++)
                {
                    if (selection->band[band].isSelectedPt[iBase][iFrame] == amdlibTRUE)
                    {
                        cpxVis.re += cpxVisTablePtr[iFrame][iBase].vis[lVis].re;
                        cpxVis.im += cpxVisTablePtr[iFrame][iBase].vis[lVis].im;
                    }
                }
                cpxVis.re /= selection->band[band].nbSelectedFrames[iBase];
                cpxVis.im /= selection->band[band].nbSelectedFrames[iBase];
                visTablePtr[iBin][iBase].vis[lVis].re=cpxVis.re;
                visTablePtr[iBin][iBase].vis[lVis].im=cpxVis.im;
                rmsVis.re = 0;
                rmsVis.im = 0;
                for (iFrame = 0; iFrame < nbFrames; iFrame++)
                {
                    if (selection->band[band].isSelectedPt[iBase][iFrame] == amdlibTRUE)
                    {
                        rmsVis.re += amdlibPow2(
                            cpxVisTablePtr[iFrame][iBase].vis[lVis].re - 
                            cpxVis.re);
                        rmsVis.im += amdlibPow2(
                            cpxVisTablePtr[iFrame][iBase].vis[lVis].im - 
                            cpxVis.im);
                    }
                }
                rmsVis.re /= selection->band[band].nbSelectedFrames[iBase];
                rmsVis.im /= selection->band[band].nbSelectedFrames[iBase];
                visTablePtr[iBin][iBase].sigma2Vis[lVis].re=rmsVis.re;
                visTablePtr[iBin][iBase].sigma2Vis[lVis].im=rmsVis.im;
            }
        }
    }
    amdlibAverageVisData_FREEALL();
    return amdlibSUCCESS;
}
#undef amdlibAverageVisData_FREEALL


/** Useful macro to free amdlibComputeDiffVis() structures */ 

#define amdlibComputeDiffVis_FREEALL()                         \
    free(cpxVisVectR); free(cpxVisVectI);                      \
    free(snrVect); free(phiVect); free(wlen);                  \
    amdlibFree2DArrayWrapping((void **)cpxVisTablePtr);        \
    amdlibFree2DArrayWrapping((void **)visTablePtr);           \
    amdlibFree2DArrayFloatWrapping(instantOpdPistonPtr);       \
    amdlibFree3DArrayComplex(cNopTable);                       \
    amdlibFree3DArrayComplex(w1);                              \
    amdlibFree3DArrayComplex(cpxVisTable);                     \
    amdlibFree3DArrayComplex(cRefTable);                       \
    amdlibFree3DArrayComplex(sigma2_cpxVisTable);              \
    amdlibFree3DArrayComplex(sigma2_cRefTable);                \
    amdlibFree3DArrayComplex(sigma2_w1);
    
/**
 * Compute differential visibility, in the way described by F. Millour et al's 
 * ("DIFFERENTIAL INTERFEROMETRY WITH THE AMBER/VLTI INSTRUMENT: DESCRIPTION, 
 * PERFORMANCES AND ILLUSTRATION", EAS Publications Series, Vol. ?, 2005)
 *
 * DiffVis are to be written in the final OI_FITS VISAMP and
 * VISPHI OI_VIS columns.
 *
 * @param instantCpxVis the bin of  instanteous correlated fluxes.
 * @param wave the table of wavelengths.
 * @param instantOpd the table of opds
 * @param band the current band
 * @param iBin index of the results to be written in the output 
 * (binned) vis structure.
 * @param selection the selection structure giving list of selected frames.
 * @param vis structure were results are stored.
 * @param errMsg error description message returned if function fails.
 *
 * @return 
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.
 */
amdlibCOMPL_STAT amdlibComputeDiffVis(/* Input */
                                      amdlibVIS              *instantCpxVis,
                                      amdlibWAVELENGTH       *wave,
                                      amdlibPISTON           *instantOpd,
                                      int                    iBin,
                                      amdlibBAND             band,
                                      amdlibSELECTION        *selection,
				                      amdlibERROR_TYPE       errorType,
                                      /* Output */
                                      amdlibVIS              *vis,
                                      amdlibERROR_MSG        errMsg)
{

    int iFrame, iSelFram; /* usually loop on number of frames */
    int nbFrames = instantCpxVis->nbFrames;
    int iBase; /*usually loop on nbBases*/
    int nbBases = instantCpxVis->nbBases;
    int lVis = 0;
    int nbLVis = instantCpxVis->nbWlen;

    amdlibVIS_TABLE_ENTRY **cpxVisTablePtr = NULL; 
    amdlibVIS_TABLE_ENTRY **visTablePtr = NULL; 
    float **instantOpdPistonPtr = NULL;
    /* We try to follow the paper's notations */
    amdlibCOMPLEX *** cNopTable = NULL; 
    amdlibCOMPLEX *** w1 = NULL; 
    amdlibCOMPLEX *** sigma2_w1 = NULL; 
    amdlibCOMPLEX *** cRefTable = NULL;
    amdlibCOMPLEX *** sigma2_cRefTable = NULL;
    amdlibCOMPLEX *** cpxVisTable = NULL;
    amdlibCOMPLEX *** sigma2_cpxVisTable = NULL;
    float  *wlen= NULL; 
    double *cpxVisVectR= NULL;
    double *cpxVisVectI= NULL;
    double *snrVect = NULL;
    double *phiVect = NULL;


    amdlibCOMPLEX phasor, cpxVis, sigma2_cpxVis, w1Avg = {0.0,0.0};
    double x;

    amdlibLogTrace("amdlibComputeDiffVis()");

    /* Allocate memory for intermediate structures, of maximum size in frames */
    cNopTable = amdlibAlloc3DArrayComplex(nbLVis, nbBases, nbFrames, errMsg);
    if (cNopTable == NULL)
    {
        amdlibComputeDiffVis_FREEALL();
        amdlibSetErrMsg("Allocation of cNopTable failed");
        return amdlibFAILURE;
    }
    w1 = amdlibAlloc3DArrayComplex(nbLVis, nbBases, nbFrames, errMsg);
    if (w1 == NULL)
    {
        amdlibComputeDiffVis_FREEALL();
        amdlibSetErrMsg("Allocation of w1 failed");
        return amdlibFAILURE;
    }
    sigma2_w1 = amdlibAlloc3DArrayComplex(nbLVis, nbBases, nbFrames, errMsg);
    if (sigma2_w1 == NULL)
    {
        amdlibComputeDiffVis_FREEALL();
        amdlibSetErrMsg("Allocation of w1 failed");
        return amdlibFAILURE;
    }
    cpxVisTable = amdlibAlloc3DArrayComplex(nbLVis, nbBases, nbFrames, errMsg);
    if (cpxVisTable == NULL)
    {
        amdlibComputeDiffVis_FREEALL();
        amdlibSetErrMsg("Allocation of cpxVisTable failed");
        return amdlibFAILURE;
    }
    sigma2_cpxVisTable = amdlibAlloc3DArrayComplex(nbLVis, nbBases, 
                                                   nbFrames, errMsg);
    if (sigma2_cpxVisTable == NULL)
    {
        amdlibComputeDiffVis_FREEALL();
        amdlibSetErrMsg("Allocation of cpxVisTableErr failed");
        return amdlibFAILURE;
    }
    cRefTable = amdlibAlloc3DArrayComplex(nbLVis, nbBases, nbFrames, errMsg);
    if (cRefTable == NULL)
    {
        amdlibComputeDiffVis_FREEALL();
        amdlibSetErrMsg("Allocation of cRefTable failed");
        return amdlibFAILURE;
    }
    sigma2_cRefTable = amdlibAlloc3DArrayComplex(nbLVis, nbBases, 
                                                 nbFrames, errMsg);
    if (sigma2_cRefTable == NULL)
    {
        amdlibComputeDiffVis_FREEALL();
        amdlibSetErrMsg("Allocation of cRefTableErr failed");
        return amdlibFAILURE;
    }
    wlen = calloc(nbLVis,sizeof(*wlen));
    if (wlen == NULL)
    {
        amdlibComputeDiffVis_FREEALL();
        amdlibSetErrMsg("Allocation of wlen failed");
        return amdlibFAILURE;
    }

    cpxVisVectR = calloc(nbFrames,sizeof(*cpxVisVectR));
    if (cpxVisVectR == NULL)
    {
        amdlibComputeDiffVis_FREEALL();
        amdlibSetErrMsg("Allocation of cpxVisVectR failed");
        return amdlibFAILURE;
    }
    cpxVisVectI = calloc(nbFrames,sizeof(*cpxVisVectI));
    if (cpxVisVectI == NULL)
    {
        amdlibComputeDiffVis_FREEALL();
        amdlibSetErrMsg("Allocation of cpxVisVectI failed");
        return amdlibFAILURE;
    }
    snrVect = calloc(nbFrames,sizeof(*snrVect));
    if ( snrVect== NULL)
    {
        amdlibComputeDiffVis_FREEALL();
        amdlibSetErrMsg("Allocation of snrVect failed");
        return amdlibFAILURE;
    }
    phiVect = calloc(nbFrames,sizeof(*phiVect));
    if ( phiVect== NULL)
    {
        amdlibComputeDiffVis_FREEALL();
        amdlibSetErrMsg("Allocation of phiVect failed");
        return amdlibFAILURE;
    }
    /* Wrap structures given as parametres */
    cpxVisTablePtr = 
        (amdlibVIS_TABLE_ENTRY **)amdlibWrap2DArray(instantCpxVis->table,
                                            nbBases, nbFrames,
                                            sizeof(amdlibVIS_TABLE_ENTRY),
                                            errMsg);
    if (cpxVisTablePtr == NULL)
    {
        amdlibComputeDiffVis_FREEALL();
        return amdlibFAILURE;
    }

    instantOpdPistonPtr = 
        amdlibWrap2DArrayFloat(instantOpd->pistonOPDArray[band],
                               instantOpd->nbBases,
                               instantOpd->nbFrames, errMsg);
    if (instantOpdPistonPtr == NULL)
    {
        amdlibSetErrMsg("Could not wrap 2D float array");
        amdlibComputeDiffVis_FREEALL();
        return amdlibFAILURE;
    }

    visTablePtr = 
        (amdlibVIS_TABLE_ENTRY **)amdlibWrap2DArray(vis->table,
                                                nbBases, vis->nbFrames,
                                                sizeof(amdlibVIS_TABLE_ENTRY),
                                                errMsg);
    if (visTablePtr == NULL)
    {
        amdlibComputeDiffVis_FREEALL();
        return amdlibFAILURE;
    }

    /* Immediately copy input cpxVis to 3D structure */

    for (iFrame = 0; iFrame < nbFrames; iFrame++)
    {
        for (iBase = 0; iBase < nbBases; iBase++)
        {
            for (lVis = 0; lVis < nbLVis; lVis++)
            {
                cpxVisTable[iFrame][iBase][lVis].re = 
                    cpxVisTablePtr[iFrame][iBase].vis[lVis].re;
                cpxVisTable[iFrame][iBase][lVis].im = 
                    cpxVisTablePtr[iFrame][iBase].vis[lVis].im;

                sigma2_cpxVisTable[iFrame][iBase][lVis].re = 
                    cpxVisTablePtr[iFrame][iBase].sigma2Vis[lVis].re;
                sigma2_cpxVisTable[iFrame][iBase][lVis].im = 
                    cpxVisTablePtr[iFrame][iBase].sigma2Vis[lVis].im;
            }
        }
    }
    /* Initialize wlen */
    for (lVis = 0; lVis < nbLVis; lVis++)
    {
        wlen[lVis] = wave->wlen[lVis];
    }

    /* Correct from Piston, 1st time */
    /* First, correct coherent flux from achromatic piston phase (eq 2.2)*/
    if (amdlibCorrect3DVisTableFromAchromaticPiston(cpxVisTable,
						    cNopTable,
						    nbFrames,
						    nbBases,
						    nbLVis,
						    wlen,
						    instantOpdPistonPtr,
						    errMsg) != amdlibSUCCESS)
    {
        return amdlibFAILURE;
    }
    
    /* Production of the reference channel: mean of all R and I parts
     * of all channels except the one considered ( eq 2.3) */
    for (iFrame = 0; iFrame < nbFrames; iFrame++)
    {
        for (iBase = 0; iBase < nbBases; iBase++)
        {
            cpxVis.re = 0.0;
            cpxVis.im = 0.0;
            sigma2_cpxVis.re = 0.0;
            sigma2_cpxVis.im = 0.0;
            /* sum all R and I */
            for (lVis = 0; lVis < nbLVis; lVis++)
            {
                cpxVis.re += cNopTable[iFrame][iBase][lVis].re;
                cpxVis.im += cNopTable[iFrame][iBase][lVis].im;

                sigma2_cpxVis.re += sigma2_cpxVisTable[iFrame][iBase][lVis].re;
                sigma2_cpxVis.im += sigma2_cpxVisTable[iFrame][iBase][lVis].im;
            }
            /* then construct Cref by substracting current R and I 
             * at that Wlen and make the arithmetic mean */
            for (lVis = 0; lVis < nbLVis; lVis++)
            {
                cRefTable[iFrame][iBase][lVis].re = 
                    (cpxVis.re - cNopTable[iFrame][iBase][lVis].re) /
                    (nbLVis - 1);
                cRefTable[iFrame][iBase][lVis].im =
                    (cpxVis.im - cNopTable[iFrame][iBase][lVis].im) /
                    (nbLVis - 1);
		
                sigma2_cRefTable[iFrame][iBase][lVis].re = 
                    (sigma2_cpxVis.re -
		     sigma2_cpxVisTable[iFrame][iBase][lVis].re) /
                    (nbLVis - 1);
                sigma2_cRefTable[iFrame][iBase][lVis].im =
                    (sigma2_cpxVis.im -
		     sigma2_cpxVisTable[iFrame][iBase][lVis].im) /
                    (nbLVis - 1);
            }
        }
    }

    /* Now the interspectrum is cNopTable*~cRefTable. Store in w1. */
    for (iBase = 0; iBase < nbBases; iBase++)
    {
        for (iFrame = 0; iFrame < nbFrames; iFrame++)
        {
            for (lVis = 0; lVis < nbLVis; lVis++)
            {
                cpxVis.re = cNopTable[iFrame][iBase][lVis].re;
                cpxVis.im = cNopTable[iFrame][iBase][lVis].im;

                phasor.re = cRefTable[iFrame][iBase][lVis].re;
                phasor.im = -1*cRefTable[iFrame][iBase][lVis].im;

                amdlibCpxMul(cpxVis,phasor,w1[iFrame][iBase][lVis]);
		
		/* Please have a look to the F. Millour thesis
		   (http://tel.archives-ouvertes.fr/tel-00134268),
		   pp.91-892 (eq. 4.55 to 4.58) */
		sigma2_w1[iFrame][iBase][lVis].re =
		    sigma2_cpxVisTable[iFrame][iBase][lVis].re *
		    amdlibPow2(cRefTable[iFrame][iBase][lVis].re) +
		    sigma2_cRefTable[iFrame][iBase][lVis].re *
		    amdlibPow2(cpxVisTable[iFrame][iBase][lVis].re) +
		    sigma2_cpxVisTable[iFrame][iBase][lVis].im *
		    amdlibPow2(cRefTable[iFrame][iBase][lVis].im) +
		    sigma2_cRefTable[iFrame][iBase][lVis].im *
		    amdlibPow2(cpxVisTable[iFrame][iBase][lVis].im);
		
		sigma2_w1[iFrame][iBase][lVis].im =
		    sigma2_cpxVisTable[iFrame][iBase][lVis].im *
		    amdlibPow2(cRefTable[iFrame][iBase][lVis].re) +
		    sigma2_cRefTable[iFrame][iBase][lVis].im *
		    amdlibPow2(cpxVisTable[iFrame][iBase][lVis].re) +
		    sigma2_cpxVisTable[iFrame][iBase][lVis].re *
		    amdlibPow2(cRefTable[iFrame][iBase][lVis].im) +
		    sigma2_cRefTable[iFrame][iBase][lVis].re *
		    amdlibPow2(cpxVisTable[iFrame][iBase][lVis].im);
            }
        }
    }

    /* Compute mean VisPhi as average of selected frames. */
    for (lVis = 0; lVis < nbLVis; lVis++)
    {
        for (iBase = 0; iBase < nbBases; iBase++)
        {
            /* The SNR vector */
            for (iFrame = 0 , iSelFram = 0 ; iFrame < nbFrames; iFrame++)
            {
                if (selection->band[band].isSelectedPt[iBase][iFrame] == amdlibTRUE)
                {
                    snrVect[iSelFram] = 1.0;
                    iSelFram++;
                }
            }
            /* The W1 vector */
            for (iFrame = 0 , iSelFram = 0 ; iFrame < nbFrames; iFrame++)
            {
                if (selection->band[band].isSelectedPt[iBase][iFrame] == amdlibTRUE)
                {
                    cpxVisVectR[iSelFram] = w1[iFrame][iBase][lVis].re;
                    cpxVisVectI[iSelFram] = w1[iFrame][iBase][lVis].im;
                    iSelFram++;
                } 
            }
            /* The Phase Herself */
            if (selection->band[band].nbSelectedFrames[iBase] != 0)
            {
                /* average re and im */
                w1Avg.re = amdlibAvgTable(selection->band[band].nbSelectedFrames[iBase],
					  cpxVisVectR  , snrVect);
                w1Avg.im = amdlibAvgTable(selection->band[band].nbSelectedFrames[iBase],
					  cpxVisVectI  , snrVect);
                /* store */
                visTablePtr[iBin][iBase].diffVisPhi[lVis] =
		    (double)(nbLVis-1)/(double)nbLVis*atan2(w1Avg.im,w1Avg.re);
            }

            /* The Error on Phi */

	    if (errorType == amdlibTHEORETICAL_ERROR)
		/* THEORETICAL ERROR, here just in case */
	    {
		/* FIXME: NOT IMPLEMENTED YET ! */
		visTablePtr[iBin][iBase].diffVisPhiErr[lVis] = 0.0;
	    }
	    else if(errorType == amdlibSTATISTICAL_ERROR)
		/* STATISTICAL ERROR */
	    {
		w1Avg.im=-1*w1Avg.im;
		for (iFrame = 0 , iSelFram = 0 ; iFrame < nbFrames; iFrame++)
		{
		    if (selection->band[band].isSelectedPt[iBase][iFrame] == amdlibTRUE)
		    {
			/* add w1*conj(w1Avg) to vector*/
			amdlibCpxMul(w1[iFrame][iBase][lVis],w1Avg,cpxVis);
			/* use phiVect to store phases */
			phiVect[iSelFram] = atan2(cpxVis.im,cpxVis.re);
			iSelFram++;
		    } 
		}
		if (selection->band[band].nbSelectedFrames[iBase] != 0)
		{
		    x = amdlibRmsTable(selection->band[band].nbSelectedFrames[iBase],
				       phiVect  , snrVect);
		    /* Err on Phi must be corrected with an abacus*/
		    visTablePtr[iBin][iBase].diffVisPhiErr[lVis] =
			amdlibAbacusErrPhi(x/sqrt(selection->band[band].nbSelectedFrames[iBase]));
		}
	    }
        }
    }




    
    /* Now for the differential vis, we use cNopTable/cRefTable. Store in w1. 
    * note this is not exactly as compliacted as in the computeDiff yorick 
    * implementation by Millour */
    for (iBase = 0; iBase < nbBases; iBase++)
    {
        for (iFrame = 0; iFrame < nbFrames; iFrame++)
        {
            for (lVis = 0; lVis < nbLVis; lVis++)
            {
                cpxVis.re = cNopTable[iFrame][iBase][lVis].re;
                cpxVis.im = cNopTable[iFrame][iBase][lVis].im;
                phasor.re = cRefTable[iFrame][iBase][lVis].re;
                phasor.im = cRefTable[iFrame][iBase][lVis].im;
                amdlibCpxDiv(cpxVis,phasor,w1[iFrame][iBase][lVis]);
            }
        }
    }
     /* Compute mean VisAmp as average of selected frames. */
    for (lVis = 0; lVis < nbLVis; lVis++)
    {
        for (iBase = 0; iBase < nbBases; iBase++)
        {
            /* The SNR vector */
            for (iFrame = 0 , iSelFram = 0 ; iFrame < nbFrames; iFrame++)
            {
                if (selection->band[band].isSelectedPt[iBase][iFrame] == amdlibTRUE)
                {
                    snrVect[iSelFram] = 1.0;
                    iSelFram++;
                }
            }
            /* The W1 vector */
            for (iFrame = 0 , iSelFram = 0 ; iFrame < nbFrames; iFrame++)
            {
                if (selection->band[band].isSelectedPt[iBase][iFrame] == amdlibTRUE)
                {
                    /* see eq 2.8 */
                    x = visTablePtr[iBin][iBase].diffVisPhi[lVis];
                    phasor.re = cos(x);
                    phasor.im = -sin(x);
                    cpxVis.re = w1[iFrame][iBase][lVis].re;
                    cpxVis.im = w1[iFrame][iBase][lVis].im;
                    cpxVisVectR[iSelFram] =
                        phasor.re * cpxVis.re - phasor.im * cpxVis.im;
                    cpxVisVectI[iSelFram] =
                        phasor.re * cpxVis.im + phasor.im * cpxVis.re;
                    iSelFram++;
                }
            }
            if (selection->band[band].nbSelectedFrames[iBase] != 0)
            {
                visTablePtr[iBin][iBase].diffVisAmp[lVis] =
                amdlibAvgTable(selection->band[band].nbSelectedFrames[iBase],
                               cpxVisVectR, snrVect);
                visTablePtr[iBin][iBase].diffVisAmpErr[lVis] =
                amdlibRmsTable(selection->band[band].nbSelectedFrames[iBase],
                               cpxVisVectR, snrVect) /
                sqrt(selection->band[band].nbSelectedFrames[iBase]);
            }
        }
    }

    amdlibComputeDiffVis_FREEALL();
    return amdlibSUCCESS;
}
#undef amdlibComputeDiffVis_FREEALL

/** Useful macro to free all dynamically allocated structures */ 
#define amdlibComputeFringeCriterion_FREEALL()            \
    amdlibFree2DArrayWrapping((void **)corrFluxTablePtr); \
    amdlibFree2DArrayWrapping((void **)visTablePtr);
/**
 * Compute fringe criterion. 
 * 
 * @param instantCorrFlux the bin of  instanteous correlated fluxes.
 * @param iBin index of the results to be written in the output (binned) vis 
 * structure.
 * @param band current band treated.
 * @param selection structure containing information relative to frame
 * selection. By default, all frames are selected.
 * @param vis structure were results are stored.
 */
void amdlibComputeFringeCriterion(/* Input */
                                  amdlibVIS       *instantCorrFlux,
                                  int             iBin,
                                  amdlibBAND      band,
                                  amdlibSELECTION *selection,
                                  /* Output */
                                  amdlibVIS       *vis)
{
    /* loop indexes and associated limits */
    int iFrame; /* usually loop on number of frames */
    int nbFrames = instantCorrFlux->nbFrames;

    int iBase; /*usually loop on nbBases*/
    int nbBases = instantCorrFlux->nbBases; 
    /* Number of Baselines in Science Data*/
    int lVis = 0;
    int nbLVis = instantCorrFlux->nbWlen;

    /* Input values of the coherent flux */
    amdlibCOMPLEX CorrFlux;
    amdlibCOMPLEX sigma2_CorrFlux;

    /* Used for the fringes detection criterion */
    double num_SNR_CorrFlux_re;
    double num_SNR_CorrFlux_im;
    double denom_SNR_CorrFlux_re;
    double denom_SNR_CorrFlux_im;
    double frgContrastSNR;
    double sqFrgContrastSNR;

    amdlibVIS_TABLE_ENTRY **corrFluxTablePtr = NULL; 
    amdlibVIS_TABLE_ENTRY **visTablePtr = NULL; 
    static amdlibERROR_MSG errMsg;

    /* Computation of the fringe Criterion */
    /* This is useful for the fringe Criterion :
     * initialisation of the photometry and the sigma2 */

    amdlibLogTrace("amdlibComputeFringeCriterion()");

    /* Wrap structures given as parametres */
    visTablePtr = 
        (amdlibVIS_TABLE_ENTRY **)amdlibWrap2DArray(vis->table,
                                            nbBases, vis->nbFrames,
                                            sizeof(amdlibVIS_TABLE_ENTRY),
                                            errMsg);
    if (visTablePtr == NULL)
    {
        amdlibComputeFringeCriterion_FREEALL();
    }
    corrFluxTablePtr = 
        (amdlibVIS_TABLE_ENTRY **)amdlibWrap2DArray(instantCorrFlux->table,
                                            nbBases, nbFrames,
                                            sizeof(amdlibVIS_TABLE_ENTRY),
                                            errMsg);
    if (corrFluxTablePtr == NULL)
    {
        amdlibComputeFringeCriterion_FREEALL();
    }

    for (iBase = 0; iBase < nbBases; iBase++)
    {
        if (selection->band[band].nbSelectedFrames[iBase] == 0)
        {
            continue;
        }
        num_SNR_CorrFlux_re   = 0.0;
        num_SNR_CorrFlux_im   = 0.0;
        denom_SNR_CorrFlux_re = 0.0;
        denom_SNR_CorrFlux_im = 0.0;

        for (iFrame = 0; iFrame < nbFrames; iFrame++)
        {
            if (selection->band[band].isSelectedPt[iBase][iFrame] == amdlibTRUE)
            {
                for (lVis = 0; lVis < nbLVis; lVis++)
                {
                    /* Get useful information for computing */
                    sigma2_CorrFlux.re = 
                        corrFluxTablePtr[iFrame][iBase].sigma2Vis[lVis].re;
                    sigma2_CorrFlux.im = 
                        corrFluxTablePtr[iFrame][iBase].sigma2Vis[lVis].im;
                    CorrFlux.re = corrFluxTablePtr[iFrame][iBase].vis[lVis].re;
                    CorrFlux.im = corrFluxTablePtr[iFrame][iBase].vis[lVis].im;

                    /* summed over wavelength */
                    num_SNR_CorrFlux_re += 
                        ((amdlibPow2(CorrFlux.re) / sigma2_CorrFlux.re) - 1) / 
                        sigma2_CorrFlux.re;

                    num_SNR_CorrFlux_im +=
                        ((amdlibPow2(CorrFlux.im) / sigma2_CorrFlux.im) - 1) /
                        sigma2_CorrFlux.im;

                    denom_SNR_CorrFlux_re += 1 / sigma2_CorrFlux.re;

                    denom_SNR_CorrFlux_im += 1 / sigma2_CorrFlux.im;
                }
            }
        }

        /* Compute fringe contrast SNR */
        sqFrgContrastSNR = num_SNR_CorrFlux_re / denom_SNR_CorrFlux_re +
            num_SNR_CorrFlux_im / denom_SNR_CorrFlux_im;
        frgContrastSNR = amdlibSignedSqrt(sqFrgContrastSNR);


        /* Store fringe contrast SNR */
        visTablePtr[iBin][iBase].bandFlag[band] = amdlibTRUE;
        visTablePtr[iBin][iBase].frgContrastSnrArray[band] = frgContrastSNR;
        visTablePtr[iBin][iBase].frgContrastSnr = frgContrastSNR;
    }

    amdlibComputeFringeCriterion_FREEALL();
}
#undef amdlibComputeFringeCriterion_FREEALL

/**
 * Compute global averaged values of squared visibilities (used as quality 
 * control).
 * 
 * @param vis2 structure used in input and also were results are stored.
 */
void amdlibAverageVis2(amdlibVIS2 *vis2)
{
    double w, w1, w2, w3, ww, ww1, ww2, ww3;
    double vis12 = 0.;
    double vis23 = 0.;
    double vis31 = 0.;
    double sigmaVis12 = 0.;
    double sigmaVis23 = 0.;
    double sigmaVis31 = 0.;
    int lVis, nbWlen = vis2->nbWlen;
    int iBin, nbBins = vis2->nbFrames;
    int nbTel = 2;
    int nbBases = vis2->nbBases;
    amdlibVIS2_TABLE_ENTRY **tablePtr = NULL;
    static amdlibERROR_MSG errMsg;

    amdlibLogTrace("amdlibAverageVis2()");

    /* Wrap input parametre */
    tablePtr = 
        (amdlibVIS2_TABLE_ENTRY **)amdlibWrap2DArray(vis2->table,
                                                nbBases, nbBins,
                                                sizeof(amdlibVIS2_TABLE_ENTRY),
                                                errMsg);
    if (tablePtr == NULL)
    {
        amdlibFree2DArrayWrapping((void **)tablePtr);
    }

    /* Initialize variables */
    if(nbBases == 3)
    {
        nbTel = 3;
    }
    vis2->vis12 = 0;
    vis2->vis23 = 0;
    vis2->vis31 = 0;
    vis2->sigmaVis12 = 0;
    vis2->sigmaVis23 = 0;
    vis2->sigmaVis31 = 0;
    ww1 = 0.;
    ww2 = 0.;
    ww3 = 0.;

#if amdlibDEBUG 
    if(nbBases == 3)    
    {
        amdlibLogTrace(" Wlen   VIS2 12 (  err   ) VIS2 23 (  err   ) "
                       "VIS2 31 (  err   )"); 
    }
    else
    {
        amdlibLogTrace(" Wlen      VIS2 (  err   )");
    }
#endif

    /* Compute mean visibility per spectral channel */
    for (lVis = 0; lVis < nbWlen; lVis++)
    {
        vis12 = 0.;
        vis23 = 0.;
        vis31 = 0.;
        sigmaVis12 = 0.;
        sigmaVis23 = 0.;
        sigmaVis31 = 0.;
        w1 = 0.;
        w2 = 0.;
        w3 = 0.;

        for (iBin = 0; iBin < nbBins; iBin++)
        {
            w = 1.0 / amdlibPow2(tablePtr[iBin][0].vis2Error[lVis]);
            w1 += w;  
            sigmaVis12 += tablePtr[iBin][0].vis2Error[lVis] * w;
            vis12 += tablePtr[iBin][0].vis2[lVis] * w; 
            if (nbTel == 3)
            {   
                w= 1.0 / amdlibPow2(tablePtr[iBin][1].vis2Error[lVis]);
                w2 += w;
                sigmaVis23 += tablePtr[iBin][1].vis2Error[lVis] * w;
                vis23 += tablePtr[iBin][1].vis2[lVis] * w; 
                w = 1.0 / amdlibPow2(tablePtr[iBin][2].vis2Error[lVis]);
                w3 += w;
                sigmaVis31 += tablePtr[iBin][2].vis2Error[lVis] * w;
                vis31 += tablePtr[iBin][2].vis2[lVis] * w; 
            }
        }
        vis12 /= w1;
        sigmaVis12 /= w1;
        if (nbBases == 3)
        {
            vis23 /= w2;
            sigmaVis23 /= w2;
            vis31 /= w3;
            sigmaVis31 /= w3;
        } 
        else  
        {
            vis23 = 0.;
            vis31 = 0.;
            sigmaVis23 = 0.;
            sigmaVis31 = 0.;
        } 
        ww = 1.0 / amdlibPow2(sigmaVis12);
        ww1 += ww;
        vis2->sigmaVis12 += sigmaVis12 * ww;
        vis2->vis12 += vis12 * ww;

#if amdlibDEBUG
        amdlibLogTrace("%d ", lVis);
        amdlibLogTrace("%8.3f(%8.03g)", vis12, sigmaVis12);
#endif
        if (nbBases == 3)
        {
            ww = 1.0 / amdlibPow2(sigmaVis23);
            ww2 += ww;
            vis2->sigmaVis23 += sigmaVis23 * ww;
            vis2->vis23 += vis23 * ww;

            ww = 1.0 / amdlibPow2(sigmaVis31);
            ww3 += ww;
            vis2->sigmaVis31 += sigmaVis31 * ww;
            vis2->vis31 += vis31 * ww;
#if amdlibDEBUG
            amdlibLogTrace("%8.3f(%8.03g)", vis23, sigmaVis23);
            amdlibLogTrace("%8.3f(%8.03g)", vis31, sigmaVis31);
#endif
        }
        else 
        {
#if amdlibDEBUG
            amdlibLogTrace("");
#endif
        }
    }

#if amdlibDEBUG
    amdlibLogTrace("");
#endif
    vis2->vis12 /= ww1;
    vis2->sigmaVis12 /= ww1;
    if (nbBases == 3)
    {
        vis2->vis23 /= ww2;
        vis2->vis31 /= ww3;
        vis2->sigmaVis23 /= ww2;
        vis2->sigmaVis31 /= ww3;
    }

    amdlibFree2DArrayWrapping((void **)tablePtr);
}


/**
 * Fills the target structure from the values in the header of the raw data. 
 * 
 * @param data pointer to raw data.
 * @param target pointer to target structure
 * 
 * @return 
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.  */
amdlibCOMPL_STAT amdlibGetOiTargetFromRawData(amdlibRAW_DATA  *data,
                                              amdlibOI_TARGET *target)
{

    int i;
    int cIndex = 0;
    char targVal[amdlibKEYW_VAL_LEN];

    amdlibLogTrace("amdlibFillOiTargetTableEntry()");

    target->element[cIndex].targetId = 1;

    /* Retrieve from Configuration Keyword values used for various headers */
    for (i=0; i < data->insCfg.nbKeywords; i++)
    {
        if (strstr(data->insCfg.keywords[i].name, "ESO OBS TARG NAME") != NULL)
        {
            strncpy(targVal, data->insCfg.keywords[i].value, 
                    amdlibKEYW_VAL_LEN);
            amdlibStripQuotes(targVal);
            strncpy(target->element[cIndex].targetName, targVal, 16);
        }
        /*Use strncmp since RA or DEC are not unique in the header*/
        if (strncmp(data->insCfg.keywords[i].name, "RA      ", 8) == 0)
        {
            sscanf(data->insCfg.keywords[i].value, "%lf", 
                   &target->element[cIndex].raEp0);
        }
        if (strncmp(data->insCfg.keywords[i].name, "DEC     ", 8) == 0)
        {
            sscanf(data->insCfg.keywords[i].value, "%lf", 
                   &target->element[cIndex].decEp0);
        }
        if (strncmp(data->insCfg.keywords[i].name, "EQUINOX ", 8) == 0)
        {
            sscanf(data->insCfg.keywords[i].value, "%f", 
                   &target->element[cIndex].equinox);
        }
        /* the following values are fixed until ESO's VLTICS writes a good 
         * OI_TARGET by itself!*/
        strncpy(target->element[cIndex].velTyp, "UNKNOWN",  8 );
        strncpy(target->element[cIndex].velDef, "OPTICAL",  8 );
        strncpy(target->element[cIndex].specTyp, "UNKNOWN",  16 );
    }
    return amdlibSUCCESS;
}


/**
 * completes the description of the vis structure
 * 
 * @param data pointer to science data.
 * @param vis the output vis structure.
 * @param errMsg error description message returned if function fails.
 * 
 * @return 
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.  */
amdlibCOMPL_STAT amdlibFillInVisTableHeader(/* Input */
                                            amdlibSCIENCE_DATA     *data,
                                            /* Output */
                                            amdlibVIS              *vis,
                                            amdlibERROR_MSG        errMsg)
{

    int nbTel = data->nbCols-1; 
    /*Number of Telescopes Used in the Science data*/

    int iBase; /* loop on nbBases */
    int nbBases = nbTel * (nbTel - 1) / 2; 
    /* Number of Baselines in Science Data*/

    /* The index of the binned visibility */ 
    int iBin;

    /* The number of visibilities stored in output table (binned) */
    int nbBins = vis->nbFrames;

    /* The nuber of frames of (unbinned) input table in 1 bin */
    int nbFramesInOneBin = data->nbFrames / nbBins;
    int iFrame;
    double time, expTime, dateObsMJD, mjdObsAtStart;

    double mjdObs = 0;
    double utcTime = 0;
    double pBase = 0;
    double pAngle = 0;
    int i;
    amdlibVIS_TABLE_ENTRY **tablePtr = NULL;

    amdlibLogTrace("amdlibFillInVisTableHeader()");

    /* Wrap output parametre */
    tablePtr = 
        (amdlibVIS_TABLE_ENTRY **)amdlibWrap2DArray(vis->table,
                                            nbBases, nbBins,
                                            sizeof(amdlibVIS_TABLE_ENTRY),
                                            errMsg);
    if (tablePtr == NULL)
    {
        amdlibFree2DArrayWrapping((void **)tablePtr);
        return amdlibFAILURE;
    }

    /* Retrieve from Configuration Keyword values used for various headers */
    for (i = 0; i < data->insCfg.nbKeywords; i++)
    {
        if (strstr(data->insCfg.keywords[i].name, "MJD-OBS") != NULL)
        {
            sscanf(data->insCfg.keywords[i].value, "%lf", &mjdObs);
        }
        if (strncmp(data->insCfg.keywords[i].name, "UTC     ",8) != 0)
        {
            sscanf(data->insCfg.keywords[i].value, "%lf", &utcTime);
        }
    }

    if (mjdObs == 0)
    {
        amdlibLogWarning("No MJD-OBS value present in file !");
    }

    /* Define date-obs as the daystart, i.e., 00:00 the same day */
    mjdObsAtStart = (int)mjdObs;
    strncpy(vis->dateObs, amdlibMJD2ISODate(mjdObsAtStart), amdlibDATE_LENGTH);
    /* Then, define time as "the mean UTC time of the measurement in seconds 
     * since 0h on DATE-OBS.", and dateObsMJD as "the mean UTC time of the 
     * measurement expressed as a modified Julian Day.". The latter is just 
     * the Time Tag, the former is the differnce in seconds between the time 
     * tag and mjdObsAtStart. Of course since there is a binning process we 
     * must average these values over the bin.
     */

    expTime = data->expTime * nbFramesInOneBin; 
    /* Set various Vis Table Entry values */
    for (iBin = 0; iBin < nbBins; iBin++)
    { 
        /*compute averaged times */
        time = 0;
        for (iFrame = iBin * nbFramesInOneBin; 
             iFrame < (iBin + 1) * nbFramesInOneBin; 
             iFrame++)
        {
            time += data->timeTag[iFrame];
        }
        dateObsMJD = time / nbFramesInOneBin;
        time = (dateObsMJD - mjdObsAtStart) * 86400.0; /*in seconds...*/

        for (iBase = 0; iBase < nbBases; iBase++)
        { 
            tablePtr[iBin][iBase].targetId = 1; 
            /*index to OI_TARGET table*/
            /* Resulting exposure time per Bin, in second */ 
            tablePtr[iBin][iBase].expTime = expTime;
            tablePtr[iBin][iBase].time = time;
            tablePtr[iBin][iBase].dateObsMJD = dateObsMJD;
            /* U,V info from Projected bases etc...*/
            pBase = data->issInfo.projectedBaseStart[iBase] +
                (iBin + 0.5) * (data->issInfo.projectedBaseEnd[iBase] - 
                            data->issInfo.projectedBaseStart[iBase]) / nbBins;
            pAngle = data->issInfo.projectedAngleStart[iBase] +
                (iBin + 0.5) * (data->issInfo.projectedAngleEnd[iBase] - 
                            data->issInfo.projectedAngleStart[iBase]) / nbBins;
            pAngle *= M_PI / 180.0;
            tablePtr[iBin][iBase].uCoord = pBase * cos(pAngle);
            tablePtr[iBin][iBase].vCoord = pBase * sin(pAngle);
            if (nbTel == 3)
            {
                tablePtr[iBin][iBase].stationIndex[0] =
                    data->issInfo.stationIndex[iBase % nbBases]; 
                tablePtr[iBin][iBase].stationIndex[1] =
                    data->issInfo.stationIndex[(iBase + 1) % nbBases];
            }
            else
            {
                tablePtr[iBin][iBase].stationIndex[0] = 
                    data->issInfo.stationIndex[0];
                tablePtr[iBin][iBase].stationIndex[1] = 
                    data->issInfo.stationIndex[1];
            }
        }
    }
    amdlibFree2DArrayWrapping((void **)tablePtr);
    return amdlibSUCCESS;
}

/**
 * completes the description of the vis2 structure
 * 
 * @param data pointer to science data.
 * @param vis2 the output vis2 structure.
 * @param errMsg error description message returned if function fails.
 * 
 * @return 
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.  */
amdlibCOMPL_STAT amdlibFillInVis2TableHeader(/* Input */
                                             amdlibSCIENCE_DATA *data,
                                             /* Output */
                                             amdlibVIS2         *vis2,
                                             amdlibERROR_MSG    errMsg)
{

    int nbTel = data->nbCols - 1; 
    /*Number of Telescopes Used in the Science data*/

    int iBase; /* loop on nbBases */
    int nbBases = nbTel * (nbTel - 1) / 2; 
    /* Number of Baselines in Science Data*/

    /* The index of the binned visibility */ 
    int iBin;

    /* The number of visibilities stored in output table (binned) */
    int nbBins = vis2->nbFrames;

    /* The nuber of frames of (unbinned) input table in 1 bin */
    int nbFramesInOneBin = data->nbFrames / nbBins;
    int iFrame;
    double time, expTime, dateObsMJD, mjdObsAtStart;

    double mjdObs = 0;
    double utcTime = 0;
    double pBase = 0;
    double pAngle = 0;
    int i;
    amdlibVIS2_TABLE_ENTRY **tablePtr = NULL;

    amdlibLogTrace("amdlibFillInVis2TableHeader()");

    /* Wrap output parametre */
    tablePtr = 
        (amdlibVIS2_TABLE_ENTRY **)amdlibWrap2DArray(vis2->table,
                                                nbBases, nbBins,
                                                sizeof(amdlibVIS2_TABLE_ENTRY),
                                                errMsg);
    if (tablePtr == NULL)
    {
        amdlibFree2DArrayWrapping((void **)tablePtr);
        return amdlibFAILURE;
    }

    /* Retrieve from Configuration Keyword values used for various headers */
    for (i = 0; i < data->insCfg.nbKeywords; i++)
    {
        if (strstr(data->insCfg.keywords[i].name, "MJD-OBS") != NULL)
        {
            sscanf(data->insCfg.keywords[i].value, "%lf", &mjdObs);
        }
        if (strncmp(data->insCfg.keywords[i].name, "UTC     ",8) != 0)
        {
            sscanf(data->insCfg.keywords[i].value, "%lf", &utcTime);
        }
    }

    if (mjdObs == 0)
    {
        amdlibLogWarning("No MJD-OBS value present in file !");
    }

    /* Define date-obs as the daystart, i.e., 00:00 the same day */
    mjdObsAtStart = (int)mjdObs;
    strncpy(vis2->dateObs, amdlibMJD2ISODate(mjdObsAtStart), amdlibDATE_LENGTH);
    /* Then, define time as "the mean UTC time of the measurement in seconds 
     * since 0h on DATE-OBS.", and dateObsMJD as "the mean UTC time of the 
     * measurement expressed as a modified Julian Day.". The latter is just 
     * the Time Tag, the former is the differnce in seconds between the time 
     * tag and mjdObsAtStart. Of course since there is a binning  process we 
     * must average these values over the bin.
     */

    expTime = data->expTime * nbFramesInOneBin; 

    /* Set various Vis2 Table Entry values */
    for (iBin = 0; iBin < nbBins; iBin++)
    { 
        /*compute averaged times */
        time = 0;
        for (iFrame = iBin * nbFramesInOneBin; 
             iFrame < (iBin + 1) * nbFramesInOneBin; 
             iFrame++)
        {
            time += data->timeTag[iFrame];
        }
        dateObsMJD = time / nbFramesInOneBin;
        time = (dateObsMJD - mjdObsAtStart) * 86400.0; /*in seconds...*/

        for (iBase = 0; iBase < nbBases; iBase++)
        { 
            tablePtr[iBin][iBase].targetId = 1; 
            /*index to OI_TARGET table*/
            tablePtr[iBin][iBase].expTime = expTime;
            tablePtr[iBin][iBase].time = time;
            tablePtr[iBin][iBase].dateObsMJD = dateObsMJD;
            pBase = data->issInfo.projectedBaseStart[iBase] + 
                (iBin + 0.5) * (data->issInfo.projectedBaseEnd[iBase] - 
                            data->issInfo.projectedBaseStart[iBase]) / nbBins;
            pAngle = data->issInfo.projectedAngleStart[iBase] + 
                (iBin + 0.5) * (data->issInfo.projectedAngleEnd[iBase] - 
                            data->issInfo.projectedAngleStart[iBase]) / nbBins;

            pAngle *= M_PI / 180.0;
            tablePtr[iBin][iBase].uCoord = pBase * cos(pAngle);
            tablePtr[iBin][iBase].vCoord = pBase * sin(pAngle);
            if (nbTel == 3)
            {
                tablePtr[iBin][iBase].stationIndex[0] = 
                    data->issInfo.stationIndex[iBase % nbBases];
                tablePtr[iBin][iBase].stationIndex[1] =
                    data->issInfo.stationIndex[(iBase + 1) % nbBases];
            }
            else
            {
                tablePtr[iBin][iBase].stationIndex[0] = 
                    data->issInfo.stationIndex[0];
                tablePtr[iBin][iBase].stationIndex[1] = 
                    data->issInfo.stationIndex[1];
            }
        }
    }

    amdlibFree2DArrayWrapping((void **)tablePtr);
    return amdlibSUCCESS;
}

/**
 * completes the description of the vis3 structure
 * 
 * @param data pointer to science data.
 * @param vis3 the output vis3 tructure
 * @param errMsg error description message returned if function fails.
 * 
 * @return 
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.  */
amdlibCOMPL_STAT amdlibFillInVis3TableHeader(/* Input */
                                             amdlibSCIENCE_DATA *data,
                                             /* Output */
                                             amdlibVIS3         *vis3,
                                             amdlibERROR_MSG    errMsg)
{

    int nbTel = data->nbCols - 1; 
    /*Number of Telescopes Used in the Science data*/

    int iClos;
    const int nbClos = 1; /* Number of Closures FIXED */

    /* The index of the binned visibility */ 
    int iBin;

    /* The number of visibilities stored in output table (binned) */
    int nbBins = vis3->nbFrames;

    /* The nuber of frames of (unbinned) input table in 1 bin */
    int nbFramesInOneBin = data->nbFrames / nbBins;
    int iFrame;
    double time, expTime, dateObsMJD, mjdObsAtStart;

    double mjdObs = 0, utcTime = 0, pBase = 0, pAngle = 0;
    double uCoord[amdlibNBASELINE],vCoord[amdlibNBASELINE];

    int i;
    /* give the impression we may have more than 3 baselines ...
     * without really doing the job! */
    const int tel1 = 0, tel2 = 1, tel3 = 2;
    amdlibVIS3_TABLE_ENTRY **tablePtr = NULL;

    amdlibLogTrace("amdlibFillInVis3TableHeader()");

    /*Return if no use. Shall we confess we have been invoked in vain?*/
    if (nbTel < 3)
    {
        return amdlibSUCCESS;
    }

    /* Wrap output parametre */
    tablePtr = 
        (amdlibVIS3_TABLE_ENTRY **)amdlibWrap2DArray(vis3->table,
                                                nbClos, nbBins,
                                                sizeof(amdlibVIS3_TABLE_ENTRY),
                                                errMsg);
    if (tablePtr == NULL)
    {
        amdlibFree2DArrayWrapping((void **)tablePtr);
        return amdlibFAILURE;
    }

    /* Retrieve from Configuration Keyword values used for various headers */
    for (i=0; i < data->insCfg.nbKeywords; i++)
    {
        if (strstr(data->insCfg.keywords[i].name,"MJD-OBS") != NULL)
        {
            sscanf(data->insCfg.keywords[i].value, "%lf", &mjdObs);
        }
        if (strncmp(data->insCfg.keywords[i].name,"UTC     ",8) != 0)
        {
            sscanf(data->insCfg.keywords[i].value, "%lf", &utcTime);
        }
    }

    if (mjdObs == 0)
    {
        amdlibLogWarning("No MJD-OBS value present in file !");
    }

    /* Define date-obs as the daystart, i.e., 00:00 the same day */
    mjdObsAtStart = (int)mjdObs;
    strncpy (vis3->dateObs, amdlibMJD2ISODate(mjdObsAtStart),amdlibDATE_LENGTH);
    /* Then, define time as "the mean UTC time of the measurement in seconds
     * since 0h on DATE-OBS.", and dateObsMJD as "the mean UTC time of the
     * measurement expressed as a modified Julian Day.". The latter is just the
     * Time Tag, the former is the differnce in seconds between the time tag and
     * mjdObsAtStart. Of course since there is a binning process we must average
     * these values over the bin.
     */

    expTime = data->expTime * nbFramesInOneBin; 

    /* Set various Vis3 Table Entry values */
    for (iBin = 0; iBin < nbBins; iBin++)
    { /* begin loop for iBin */
        /*compute averaged times */
        time = 0;
        for (iFrame = iBin * nbFramesInOneBin; 
             iFrame < (iBin+1) * nbFramesInOneBin; iFrame++)
        {
            time += data->timeTag[iFrame];
        }
        dateObsMJD = time / nbFramesInOneBin;
        time = (dateObsMJD - mjdObsAtStart) * 86400.0; /*in seconds...*/

        /*Compute right baseline lengths */
        pBase = data->issInfo.projectedBaseStart[0] + 
            (iBin + 0.5) * (data->issInfo.projectedBaseEnd[0] - 
                            data->issInfo.projectedBaseStart[0]) / nbBins;
        pAngle = data->issInfo.projectedAngleStart[0] + 
            (iBin + 0.5) * (data->issInfo.projectedAngleEnd[0] - 
                            data->issInfo.projectedAngleStart[0]) / nbBins;
        pAngle *= M_PI / 180.0;
        uCoord[0] = pBase * cos(pAngle);
        vCoord[0] = pBase * sin(pAngle);

        pBase = data->issInfo.projectedBaseStart[1] + 
            (iBin + 0.5) * (data->issInfo.projectedBaseEnd[1] - 
                            data->issInfo.projectedBaseStart[1]) / nbBins;
        pAngle = data->issInfo.projectedAngleStart[1] +
            (iBin + 0.5) * (data->issInfo.projectedAngleEnd[1] - 
                            data->issInfo.projectedAngleStart[1]) / nbBins;
        pAngle *= M_PI / 180.0;
        uCoord[1] = pBase * cos(pAngle);
        vCoord[1] = pBase * sin(pAngle);

        for (iClos = 0; iClos < nbClos; iClos++)
        {
            tablePtr[iBin][iClos].targetId = 1; /*index to OI_TARGET table*/
            tablePtr[iBin][iClos].expTime = expTime;
            tablePtr[iBin][iClos].time = time;
            tablePtr[iBin][iClos].dateObsMJD = dateObsMJD;
            tablePtr[iBin][iClos].u1Coord = uCoord[0];
            tablePtr[iBin][iClos].v1Coord = vCoord[0];
            tablePtr[iBin][iClos].u2Coord = uCoord[1];
            tablePtr[iBin][iClos].v2Coord = vCoord[1];
            /*this code could possibly be extended to more telescopes!*/
            tablePtr[iBin][iClos].stationIndex[0] = 
                data->issInfo.stationIndex[tel1];
            tablePtr[iBin][iClos].stationIndex[1] = 
                data->issInfo.stationIndex[tel2];
            tablePtr[iBin][iClos].stationIndex[2] = 
                data->issInfo.stationIndex[tel3];
        }
    }

    amdlibFree2DArrayWrapping((void **)tablePtr);
    return amdlibSUCCESS;
}




